# Copyright (c) 2014-2016, ConfigSpace developers
# Matthias Feurer
# Katharina Eggensperger
# and others (see commit history).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from collections import OrderedDict
from itertools import product
import json
import sys
import unittest

import numpy as np

from ConfigSpace import ConfigurationSpace, \
    Configuration, CategoricalHyperparameter, UniformIntegerHyperparameter, \
    Constant, EqualsCondition, NotEqualsCondition, InCondition, \
    AndConjunction, OrConjunction, ForbiddenEqualsClause, \
    ForbiddenAndConjunction, UniformFloatHyperparameter
from ConfigSpace.hyperparameters import NormalFloatHyperparameter
from ConfigSpace.exceptions import ForbiddenValueError


def byteify(input):
    print(sys.version)
    if sys.version_info >= (3, 0):
        return input

    # From http://stackoverflow.com/a/13105359/4636294
    if isinstance(input, dict):
        return {byteify(key): byteify(value)
                for key, value in input.iteritems()}
    elif isinstance(input, list):
        return [byteify(element) for element in input]
    elif isinstance(input, unicode):
        return input.encode('utf-8')
    else:
        return input


class TestConfigurationSpace(unittest.TestCase):
    # TODO generalize a few simple configuration spaces which are used over
    # and over again throughout this test suite
    # TODO make sure that every function here tests one aspect of the
    # configuration space object!
    def test_add_hyperparameter(self):
        cs = ConfigurationSpace()
        hp = UniformIntegerHyperparameter("name", 0, 10)
        cs.add_hyperparameter(hp)

    def test_add_non_hyperparameter(self):
        cs = ConfigurationSpace()
        non_hp = unittest.TestSuite()
        self.assertRaises(TypeError, cs.add_hyperparameter, non_hp)

    def test_add_hyperparameters_with_equal_names(self):
        cs = ConfigurationSpace()
        hp = UniformIntegerHyperparameter("name", 0, 10)
        cs.add_hyperparameter(hp)
        self.assertRaisesRegexp(ValueError,
                                "Hyperparameter 'name' is already in the "
                                "configuration space.",
                                cs.add_hyperparameter, hp)

    def test_illegal_default_configuration(self):
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("loss", ["l1", "l2"], default_value='l1')
        hp2 = CategoricalHyperparameter("penalty", ["l1", "l2"], default_value='l1')
        cs.add_hyperparameter(hp1)
        cs.add_hyperparameter(hp2)
        forb1 = ForbiddenEqualsClause(hp1, "l1")
        forb2 = ForbiddenEqualsClause(hp2, "l1")
        forb3 = ForbiddenAndConjunction(forb1, forb2)
        # cs.add_forbidden_clause(forb3)
        self.assertRaisesRegexp(ValueError, "Given vector violates forbidden clause \(Forbidden: loss == \'l1\' && "
            "Forbidden: penalty == \'l1\'\)", cs.add_forbidden_clause, forb3)

    def test_meta_data_stored(self):
        meta_data = {'additional': 'meta-data',
                     'useful': 'for integrations',
                     'input_id': 42}
        cs = ConfigurationSpace(meta=dict(meta_data))
        self.assertEqual(cs.meta, meta_data)

    def test_add_non_condition(self):
        cs = ConfigurationSpace()
        non_cond = unittest.TestSuite()
        self.assertRaises(TypeError, cs.add_condition, non_cond)

    def test_hyperparameters_with_valid_condition(self):
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        cs.add_hyperparameter(hp1)
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cs.add_hyperparameter(hp2)
        cond = EqualsCondition(hp2, hp1, 0)
        cs.add_condition(cond)
        self.assertEqual(len(cs._hyperparameters), 2)

    def test_condition_without_added_hyperparameters(self):
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cond = EqualsCondition(hp2, hp1, 0)
        self.assertRaisesRegexp(ValueError, "Child hyperparameter 'child' not "
                                "in configuration space.", cs.add_condition,
                                cond)
        cs.add_hyperparameter(hp1)
        self.assertRaisesRegexp(ValueError, "Child hyperparameter 'child' not "
                                "in configuration space.", cs.add_condition,
                                cond)

        # Test also the parent hyperparameter
        cs2 = ConfigurationSpace()
        cs2.add_hyperparameter(hp2)
        self.assertRaisesRegexp(ValueError, "Parent hyperparameter 'parent' "
                                "not in configuration space.",
                                cs2.add_condition, cond)

    def test_condition_with_cycles(self):
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        cs.add_hyperparameter(hp1)
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cs.add_hyperparameter(hp2)
        cond1 = EqualsCondition(hp2, hp1, 0)
        cs.add_condition(cond1)
        cond2 = EqualsCondition(hp1, hp2, 0)
        self.assertRaisesRegexp(ValueError, "Hyperparameter configuration "
                                "contains a cycle \[\['child', 'parent'\]\]",
                                cs.add_condition, cond2)

    def test_add_conjunction(self):
        hp1 = CategoricalHyperparameter("input1", [0, 1])
        hp2 = CategoricalHyperparameter("input2", [0, 1])
        hp3 = CategoricalHyperparameter("input3", [0, 1])
        hp4 = Constant("And", "True")

        cond1 = EqualsCondition(hp4, hp1, 1)
        cond2 = EqualsCondition(hp4, hp2, 1)
        cond3 = EqualsCondition(hp4, hp3, 1)

        andconj1 = AndConjunction(cond1, cond2, cond3)

        cs = ConfigurationSpace()
        cs.add_hyperparameter(hp1)
        cs.add_hyperparameter(hp2)
        cs.add_hyperparameter(hp3)
        cs.add_hyperparameter(hp4)

        cs.add_condition(andconj1)
        self.assertNotIn(hp4, cs.get_all_unconditional_hyperparameters())

    def test_add_second_condition_wo_conjunction(self):
        hp1 = CategoricalHyperparameter("input1", [0, 1])
        hp2 = CategoricalHyperparameter("input2", [0, 1])
        hp3 = Constant("And", "True")

        cond1 = EqualsCondition(hp3, hp1, 1)
        cond2 = EqualsCondition(hp3, hp2, 1)

        cs = ConfigurationSpace()
        cs.add_hyperparameter(hp1)
        cs.add_hyperparameter(hp2)
        cs.add_hyperparameter(hp3)

        cs.add_condition(cond1)
        self.assertRaisesRegexp(ValueError,
                                "Adding a second condition \(different\) for a "
                                "hyperparameter is ambigouos and "
                                "therefore forbidden. Add a conjunction "
                                "instead!",
                                cs.add_condition, cond2)

    def test_add_forbidden_clause(self):
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("input1", [0, 1])
        cs.add_hyperparameter(hp1)
        forb = ForbiddenEqualsClause(hp1, 1)
        # TODO add checking whether a forbidden clause makes sense at all
        cs.add_forbidden_clause(forb)
        # TODO add something to properly retrieve the forbidden clauses
        self.assertEqual(str(cs), "Configuration space object:\n  "
                                  "Hyperparameters:\n    input1, "
                                  "Type: Categorical, Choices: {0, 1}, "
                                  "Default: 0\n"
                                  "  Forbidden Clauses:\n"
                                  "    Forbidden: input1 == 1\n")

    def test_add_configuration_space(self):
        cs = ConfigurationSpace()
        hp1 = cs.add_hyperparameter(CategoricalHyperparameter("input1", [0, 1]))
        forb1 = cs.add_forbidden_clause(ForbiddenEqualsClause(hp1, 1))
        hp2 = cs.add_hyperparameter(UniformIntegerHyperparameter("child", 0, 10))
        cond = cs.add_condition(EqualsCondition(hp2, hp1, 0))
        cs2 = ConfigurationSpace()
        cs2.add_configuration_space('prefix', cs, delimiter='__')
        self.assertEqual(str(cs2), '''Configuration space object:
  Hyperparameters:
    prefix__child, Type: UniformInteger, Range: [0, 10], Default: 5
    prefix__input1, Type: Categorical, Choices: {0, 1}, Default: 0
  Conditions:
    prefix__child | prefix__input1 == 0
  Forbidden Clauses:
    Forbidden: prefix__input1 == 1
''')

    def test_add_configuration_space_conjunctions(self):
        cs1 = ConfigurationSpace()
        cs2 = ConfigurationSpace()

        hp1 = cs1.add_hyperparameter(CategoricalHyperparameter("input1", [0, 1]))
        hp2 = cs1.add_hyperparameter(CategoricalHyperparameter("input2", [0, 1]))
        hp3 = cs1.add_hyperparameter(UniformIntegerHyperparameter("child1", 0, 10))
        hp4 = cs1.add_hyperparameter(UniformIntegerHyperparameter("child2", 0, 10))

        cond1 = EqualsCondition(hp2, hp3, 0)
        cond2 = EqualsCondition(hp1, hp3, 5)
        cond3 = EqualsCondition(hp1, hp4, 1)
        andCond = AndConjunction(cond2, cond3)

        cs1.add_conditions([cond1, andCond])
        cs2.add_configuration_space(prefix='test', configuration_space=cs1)

        self.assertEqual(str(cs2).count('test:'), 10)
        # Check that they're equal except for the "test:" prefix
        self.assertEqual(str(cs1), str(cs2).replace('test:', ''))

    def test_add_conditions(self):
        cs1 = ConfigurationSpace()
        cs2 = ConfigurationSpace()

        hp1 = cs1.add_hyperparameter(CategoricalHyperparameter("input1", [0, 1]))
        cs2.add_hyperparameter(hp1)
        hp2 = cs1.add_hyperparameter(CategoricalHyperparameter("input2", [0, 1]))
        cs2.add_hyperparameter(hp2)
        hp3 = cs1.add_hyperparameter(UniformIntegerHyperparameter("child1", 0, 10))
        cs2.add_hyperparameter(hp3)
        hp4 = cs1.add_hyperparameter(UniformIntegerHyperparameter("child2", 0, 10))
        cs2.add_hyperparameter(hp4)

        cond1 = EqualsCondition(hp2, hp3, 0)
        cond2 = EqualsCondition(hp1, hp3, 5)
        cond3 = EqualsCondition(hp1, hp4, 1)
        andCond = AndConjunction(cond2, cond3)

        cs1.add_conditions([cond1, andCond])
        cs2.add_condition(cond1)
        cs2.add_condition(andCond)

        self.assertEqual(str(cs1), str(cs2))

    def test_get_hyperparamforbidden_clauseseters(self):
        cs = ConfigurationSpace()
        self.assertEqual(0, len(cs.get_hyperparameters()))
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        cs.add_hyperparameter(hp1)
        self.assertEqual([hp1], cs.get_hyperparameters())
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cs.add_hyperparameter(hp2)
        cond1 = EqualsCondition(hp2, hp1, 1)
        cs.add_condition(cond1)
        self.assertEqual([hp1, hp2], cs.get_hyperparameters())
        # TODO: I need more tests for the topological sort!
        self.assertEqual([hp1, hp2], cs.get_hyperparameters())

    def test_get_hyperparameters_topological_sort_simple(self):
        for iteration in range(10):
            cs = ConfigurationSpace()
            hp1 = CategoricalHyperparameter("parent", [0, 1])
            cs.add_hyperparameter(hp1)
            hp2 = UniformIntegerHyperparameter("child", 0, 10)
            cs.add_hyperparameter(hp2)
            cond1 = EqualsCondition(hp2, hp1, 0)
            cs.add_condition(cond1)
            # This automatically checks the configuration!
            Configuration(cs, dict(parent=0, child=5))

    def test_get_hyperparameters_topological_sort(self):
            # and now for something more complicated
            cs = ConfigurationSpace()
            hp1 = CategoricalHyperparameter("input1", [0, 1])
            hp2 = CategoricalHyperparameter("input2", [0, 1])
            hp3 = CategoricalHyperparameter("input3", [0, 1])
            hp4 = CategoricalHyperparameter("input4", [0, 1])
            hp5 = CategoricalHyperparameter("input5", [0, 1])
            hp6 = Constant("AND", "True")
            # More top-level hyperparameters
            hp7 = CategoricalHyperparameter("input7", [0, 1])
            # Somewhat shuffled
            hyperparameters = [hp1, hp2, hp3, hp4, hp5, hp6, hp7]

            for hp in hyperparameters:
                cs.add_hyperparameter(hp)

            cond1 = EqualsCondition(hp6, hp1, 1)
            cond2 = NotEqualsCondition(hp6, hp2, 1)
            cond3 = InCondition(hp6, hp3, [1])
            cond4 = EqualsCondition(hp5, hp3, 1)
            cond5 = EqualsCondition(hp4, hp5, 1)
            cond6 = EqualsCondition(hp6, hp4, 1)
            cond7 = EqualsCondition(hp6, hp5, 1)

            conj1 = AndConjunction(cond1, cond2)
            conj2 = OrConjunction(conj1, cond3)
            conj3 = AndConjunction(conj2, cond6, cond7)

            cs.add_condition(cond4)
            hps = cs.get_hyperparameters()
            # AND is moved to the front because of alphabetical sorting
            for hp, idx in zip(hyperparameters, [1, 2, 3, 4, 6, 0, 5]):
                self.assertEqual(hps.index(hp), idx)
                self.assertEqual(cs._hyperparameter_idx[hp.name], idx)
                self.assertEqual(cs._idx_to_hyperparameter[idx], hp.name)

            cs.add_condition(cond5)
            hps = cs.get_hyperparameters()
            for hp, idx in zip(hyperparameters, [1, 2, 3, 6, 5, 0, 4]):
                self.assertEqual(hps.index(hp), idx)
                self.assertEqual(cs._hyperparameter_idx[hp.name], idx)
                self.assertEqual(cs._idx_to_hyperparameter[idx], hp.name)

            cs.add_condition(conj3)
            hps = cs.get_hyperparameters()
            print(hps, hyperparameters)
            for hp, idx in zip(hyperparameters, [0, 1, 2, 5, 4, 6, 3]):
                print(hp, idx)
                self.assertEqual(hps.index(hp), idx)
                self.assertEqual(cs._hyperparameter_idx[hp.name], idx)
            self.assertEqual(cs._idx_to_hyperparameter[idx], hp.name)

    def test_get_hyperparameter(self):
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        cs.add_hyperparameter(hp1)
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cs.add_hyperparameter(hp2)

        retval = cs.get_hyperparameter("parent")
        self.assertEqual(hp1, retval)
        retval = cs.get_hyperparameter("child")
        self.assertEqual(hp2, retval)
        self.assertRaises(KeyError, cs.get_hyperparameter, "grandfather")

    def test_get_conditions(self):
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        cs.add_hyperparameter(hp1)
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cs.add_hyperparameter(hp2)
        self.assertEqual([], cs.get_conditions())
        cond1 = EqualsCondition(hp2, hp1, 0)
        cs.add_condition(cond1)
        self.assertEqual([cond1], cs.get_conditions())

    def test_get_parent_and_chil_conditions_of(self):
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        cs.add_hyperparameter(hp1)
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cs.add_hyperparameter(hp2)
        cond1 = EqualsCondition(hp2, hp1, 0)
        cs.add_condition(cond1)

        self.assertEqual([cond1], cs.get_parent_conditions_of(hp2.name))
        self.assertEqual([cond1], cs.get_parent_conditions_of(hp2))
        self.assertEqual([cond1], cs.get_child_conditions_of(hp1.name))
        self.assertEqual([cond1], cs.get_child_conditions_of(hp1))

        self.assertRaisesRegexp(KeyError,
                                "Hyperparameter 'Foo' does not exist in this "
                                "configuration space.", cs.get_parents_of,
                                "Foo")
        self.assertRaisesRegexp(KeyError,
                                "Hyperparameter 'Foo' does not exist in this "
                                "configuration space.", cs.get_children_of,
                                "Foo")

    def test_get_parent_and_children_of(self):
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        cs.add_hyperparameter(hp1)
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cs.add_hyperparameter(hp2)
        cond1 = EqualsCondition(hp2, hp1, 0)
        cs.add_condition(cond1)

        self.assertEqual([hp1], cs.get_parents_of(hp2.name))
        self.assertEqual([hp1], cs.get_parents_of(hp2))
        self.assertEqual([hp2], cs.get_children_of(hp1.name))
        self.assertEqual([hp2], cs.get_children_of(hp1))

        self.assertRaisesRegexp(KeyError,
                                "Hyperparameter 'Foo' does not exist in this "
                                "configuration space.", cs.get_parents_of,
                                "Foo")
        self.assertRaisesRegexp(KeyError,
                                "Hyperparameter 'Foo' does not exist in this "
                                "configuration space.", cs.get_children_of,
                                "Foo")

    def test_check_configuration_input_checking(self):
        cs = ConfigurationSpace()
        self.assertRaisesRegexp(TypeError, "The method check_configuration must be called "
                                            "with an instance of %s. "
                                            "Your input was of type %s"% (Configuration, type("String")),
                                cs.check_configuration, "String")
        # For the check configuration method with vector representation
        self.assertRaisesRegexp(TypeError, "The method check_configuration must"
                                           " be called with an instance of "
                                           "np.ndarray Your input was of type %s" % (type("String")),
                                cs.check_configuration_vector_representation, "String")

    def test_check_configuration(self):
        # TODO this is only a smoke test
        # TODO actually, this rather tests the evaluate methods in the
        # conditions module!
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        cs.add_hyperparameter(hp1)
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cs.add_hyperparameter(hp2)
        cond1 = EqualsCondition(hp2, hp1, 0)
        cs.add_condition(cond1)
        # This automatically checks the configuration!
        Configuration(cs, dict(parent=0, child=5))

        # and now for something more complicated
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("input1", [0, 1])
        cs.add_hyperparameter(hp1)
        hp2 = CategoricalHyperparameter("input2", [0, 1])
        cs.add_hyperparameter(hp2)
        hp3 = CategoricalHyperparameter("input3", [0, 1])
        cs.add_hyperparameter(hp3)
        hp4 = CategoricalHyperparameter("input4", [0, 1])
        cs.add_hyperparameter(hp4)
        hp5 = CategoricalHyperparameter("input5", [0, 1])
        cs.add_hyperparameter(hp5)
        hp6 = Constant("AND", "True")
        cs.add_hyperparameter(hp6)

        cond1 = EqualsCondition(hp6, hp1, 1)
        cond2 = NotEqualsCondition(hp6, hp2, 1)
        cond3 = InCondition(hp6, hp3, [1])
        cond4 = EqualsCondition(hp6, hp4, 1)
        cond5 = EqualsCondition(hp6, hp5, 1)

        conj1 = AndConjunction(cond1, cond2)
        conj2 = OrConjunction(conj1, cond3)
        conj3 = AndConjunction(conj2, cond4, cond5)
        cs.add_condition(conj3)

        expected_outcomes = [False, False, False, False, False,
                             False, False, True, False, False,
                             False, False, False, False, False,
                             True, False, False, False, True,
                             False, False, False, True, False,
                             False, False, False, False, False,
                             False, True]

        for idx, values in enumerate(product([0, 1], repeat=5)):
            # The hyperparameters aren't sorted, but the test assumes them to
            #  be sorted.
            hyperparameters = sorted(cs.get_hyperparameters(),
                                     key=lambda t: t.name)
            instantiations = {hyperparameters[jdx+1].name: values[jdx]
                              for jdx in range(len(values))}

            evaluation = conj3.evaluate(instantiations)
            self.assertEqual(expected_outcomes[idx], evaluation)

            if evaluation == False:
                self.assertRaisesRegexp(ValueError,
                                        "Inactive hyperparameter 'AND' must "
                                        "not be specified, but has the vector value: "
                                        "'0.0'.",
                                        Configuration, cs, values={
                                        "input1": values[0],
                                        "input2": values[1],
                                        "input3": values[2],
                                        "input4": values[3],
                                        "input5": values[4],
                                        "AND": "True"})
            else:
                Configuration(cs, values={"input1": values[0],
                                          "input2": values[1],
                                          "input3": values[2],
                                          "input4": values[3],
                                          "input5": values[4],
                                          "AND": "True"})

    def test_check_configuration2(self):
        # Test that hyperparameters which are not active must not be set and
        # that evaluating forbidden clauses does not choke on missing
        # hyperparameters
        cs = ConfigurationSpace()
        classifier = CategoricalHyperparameter("classifier",
            ["k_nearest_neighbors", "extra_trees"])
        metric = CategoricalHyperparameter("metric", ["minkowski", "other"])
        p = CategoricalHyperparameter("k_nearest_neighbors:p", [1, 2])
        metric_depends_on_classifier = EqualsCondition(metric, classifier,
                                                       "k_nearest_neighbors")
        p_depends_on_metric = EqualsCondition(p, metric, "minkowski")
        cs.add_hyperparameter(metric)
        cs.add_hyperparameter(p)
        cs.add_hyperparameter(classifier)
        cs.add_condition(metric_depends_on_classifier)
        cs.add_condition(p_depends_on_metric)

        forbidden = ForbiddenEqualsClause(metric, "other")
        cs.add_forbidden_clause(forbidden)

        configuration = Configuration(cs, dict(classifier="extra_trees"))

        # check backward compatibility with checking configurations instead of vectors
        cs.check_configuration(configuration)

    def test_check_forbidden_with_sampled_vector_configuration(self):
        cs = ConfigurationSpace()
        metric = CategoricalHyperparameter("metric", ["minkowski", "other"])
        cs.add_hyperparameter(metric)

        forbidden = ForbiddenEqualsClause(metric, "other")
        cs.add_forbidden_clause(forbidden)
        configuration = Configuration(cs,
            vector=np.ones(1, dtype=float))
        self.assertRaisesRegexp(ValueError, "violates forbidden clause",
                                cs._check_forbidden, configuration.get_array())

    def test_eq(self):
        # Compare empty configuration spaces
        cs1 = ConfigurationSpace()
        cs2 = ConfigurationSpace()
        self.assertEqual(cs1, cs2)

        # Compare to something which isn't a configuration space
        self.assertTrue(not (cs1 == "ConfigurationSpace"))

        # Compare to equal configuration spaces
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        hp3 = UniformIntegerHyperparameter("friend", 0, 5)
        cond1 = EqualsCondition(hp2, hp1, 0)
        cs1.add_hyperparameter(hp1)
        cs1.add_hyperparameter(hp2)
        cs1.add_condition(cond1)
        cs2.add_hyperparameter(hp1)
        cs2.add_hyperparameter(hp2)
        cs2.add_condition(cond1)
        self.assertEqual(cs1, cs2)
        cs1.add_hyperparameter(hp3)
        self.assertFalse(cs1 == cs2)

    def test_neq(self):
        cs1 = ConfigurationSpace()
        self.assertNotEqual(cs1, "ConfigurationSpace")

    def test_repr(self):
        cs1 = ConfigurationSpace()
        retval = cs1.__str__()
        self.assertEqual("Configuration space object:\n  Hyperparameters:\n",
                         retval)

        hp1 = CategoricalHyperparameter("parent", [0, 1])
        cs1.add_hyperparameter(hp1)
        retval = cs1.__str__()
        self.assertEqual("Configuration space object:\n  Hyperparameters:\n"
                         "    %s\n" % str(hp1), retval)

        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cond1 = EqualsCondition(hp2, hp1, 0)
        cs1.add_hyperparameter(hp2)
        cs1.add_condition(cond1)
        retval = cs1.__str__()
        self.assertEqual("Configuration space object:\n  Hyperparameters:\n"
                         "    %s\n    %s\n  Conditions:\n    %s\n" %
                         (str(hp2), str(hp1), str(cond1)), retval)

    def test_sample_configuration(self):
        cs = ConfigurationSpace()
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        cs.add_hyperparameter(hp1)
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cs.add_hyperparameter(hp2)
        cond1 = EqualsCondition(hp2, hp1, 0)
        cs.add_condition(cond1)
        # This automatically checks the configuration!
        Configuration(cs, dict(parent=0, child=5))

        # and now for something more complicated
        cs = ConfigurationSpace(seed=1)
        hp1 = CategoricalHyperparameter("input1", [0, 1])
        cs.add_hyperparameter(hp1)
        hp2 = CategoricalHyperparameter("input2", [0, 1])
        cs.add_hyperparameter(hp2)
        hp3 = CategoricalHyperparameter("input3", [0, 1])
        cs.add_hyperparameter(hp3)
        hp4 = CategoricalHyperparameter("input4", [0, 1])
        cs.add_hyperparameter(hp4)
        hp5 = CategoricalHyperparameter("input5", [0, 1])
        cs.add_hyperparameter(hp5)
        hp6 = Constant("AND", "True")
        cs.add_hyperparameter(hp6)

        cond1 = EqualsCondition(hp6, hp1, 1)
        cond2 = NotEqualsCondition(hp6, hp2, 1)
        cond3 = InCondition(hp6, hp3, [1])
        cond4 = EqualsCondition(hp5, hp3, 1)
        cond5 = EqualsCondition(hp4, hp5, 1)
        cond6 = EqualsCondition(hp6, hp4, 1)
        cond7 = EqualsCondition(hp6, hp5, 1)

        conj1 = AndConjunction(cond1, cond2)
        conj2 = OrConjunction(conj1, cond3)
        conj3 = AndConjunction(conj2, cond6, cond7)
        cs.add_condition(cond4)
        cs.add_condition(cond5)
        cs.add_condition(conj3)

        samples = []
        for i in range(5):
            cs.seed(1)
            samples.append([])
            for j in range(100):
                sample = cs.sample_configuration()
                samples[-1].append(sample)

            if i > 0:
                for j in range(100):
                    self.assertEqual(samples[-1][j], samples[-2][j])

    def test_sample_wrong_argument(self):
        cs = ConfigurationSpace()
        self.assertRaisesRegex(TypeError,
                               "Argument size must be of type int, but is "
                               "<class 'float'>", cs.sample_configuration, 1.2)

    def test_sample_no_configuration(self):
        cs = ConfigurationSpace()
        rval = cs.sample_configuration(size=0)
        self.assertEqual(len(rval), 0)


class ConfigurationTest(unittest.TestCase):
    def setUp(self):
        cs = ConfigurationSpace()
        hp1 = cs.add_hyperparameter(CategoricalHyperparameter("parent", [0, 1]))
        hp2 = cs.add_hyperparameter(
            UniformIntegerHyperparameter("child", 0, 10))
        hp3 = cs.add_hyperparameter(
            UniformIntegerHyperparameter("friend", 0, 5))
        self.cs = cs

    def test_wrong_init(self):
        self.assertRaisesRegexp(ValueError,
                               'Configuration neither specified as dictionary '
                               'or vector.', Configuration, self.cs)

        self.assertRaisesRegexp(ValueError,
                               'Configuration specified both as dictionary and '
                               'vector, can only do one.', Configuration,
                               self.cs, values={}, vector=np.zeros((3, )))

    def test_init_with_values(self):
        c1 = Configuration(self.cs, values={'parent': 1,
                                            'child': 2,
                                            'friend': 3})
        # Pay attention that the vector does not necessarily has an intuitive
        #  sorting!
        # Values are a little bit higher than one would expect because,
        # an integer range of [0,10] is transformed to [-0.499,10.499].
        vector_values = {'parent': 1,
                         'child': 0.22727223140405708,
                         'friend': 0.583333611112037}
        vector = [None] * 3
        for name in self.cs._hyperparameter_idx:
            vector[self.cs._hyperparameter_idx[name]] = vector_values[name]
        c2 = Configuration(self.cs, vector=vector)
        # This tests
        # a) that the vector representation of both are the same
        # b) that the dictionary representation of both are the same
        self.assertEqual(c1, c2)

    def test_uniformfloat_transform(self):
        """This checks whether a value sampled through the configuration
        space (it does not happend when the variable is sampled alone) stays
        equal when it is serialized via JSON and the deserialized again."""

        cs = ConfigurationSpace()
        a = cs.add_hyperparameter(UniformFloatHyperparameter('a', -5, 10))
        b = cs.add_hyperparameter(NormalFloatHyperparameter('b', 1, 2,
                                                            log=True))
        for i in range(100):
            config = cs.sample_configuration()
            value = OrderedDict(sorted(config.get_dictionary().items()))
            string = json.dumps(value)
            saved_value = json.loads(string)
            saved_value = OrderedDict(sorted(byteify(saved_value).items()))
            self.assertEqual(repr(value), repr(saved_value))

        # Next, test whether the truncation also works when initializing the
        # Configuration with a dictionary
        for i in range(100):
            rs = np.random.RandomState(1)
            value_a = a.sample(rs)
            value_b = b.sample(rs)
            values_dict = {'a': value_a, 'b': value_b}
            config = Configuration(cs, values=values_dict)
            string = json.dumps(config.get_dictionary())
            saved_value = json.loads(string)
            saved_value = byteify(saved_value)
            self.assertEqual(values_dict, saved_value)
            
    def test_setitem(self):
        '''
        Checks overriding a sampled configuration
        '''
        pcs = ConfigurationSpace()
        pcs.add_hyperparameter(UniformIntegerHyperparameter('x0', 1, 5, default_value=1))
        x1 = pcs.add_hyperparameter(CategoricalHyperparameter('x1', ['ab', 'bc', 'cd', 'de'], default_value='ab'))

        # Condition
        x2 = pcs.add_hyperparameter(CategoricalHyperparameter('x2', [1, 2]))
        pcs.add_condition(EqualsCondition(x2, x1, 'ab'))

        # Forbidden
        x3 = pcs.add_hyperparameter(CategoricalHyperparameter('x3', [1, 2]))
        pcs.add_forbidden_clause(ForbiddenEqualsClause(x3, 2))
        
        conf = pcs.get_default_configuration()
        
        # failed because it's a invalid configuration
        with self.assertRaisesRegex(ValueError, "Illegal value '0' for hyperparameter x0"):
            conf['x0'] = 0
        
        # failed because the variable didn't exists
        with self.assertRaisesRegex(KeyError, "Hyperparameter 'x_0' does not exist in this configuration space."):
            conf['x_0'] = 1

        # failed because forbidden clause is violated
        with self.assertRaisesRegex(ForbiddenValueError, "Given vector violates forbidden clause Forbidden: x3 == 2"):
            conf['x3'] = 2
        self.assertEqual(conf['x3'], 1)
        
        # successful operation 1
        x0_old = conf['x0']
        if x0_old == 1:
            conf['x0'] = 2
        else:
            conf['x0'] = 1
        x0_new = conf['x0']
        self.assertNotEqual(x0_old, x0_new)
        pcs._check_configuration_rigorous(conf)
        self.assertEqual(conf['x2'], 1)
        
        # successful operation 2
        x1_old = conf['x1']
        if x1_old == 'ab':
            conf['x1'] = 'cd'
        else:
            conf['x1'] = 'ab'
        x1_new = conf['x1']
        self.assertNotEqual(x1_old, x1_new)
        pcs._check_configuration_rigorous(conf)
        self.assertRaises(KeyError, conf.__getitem__, 'x2')

    def test_setting_illegal_value(self):
        cs = ConfigurationSpace()
        cs.add_hyperparameter(UniformFloatHyperparameter('x', 0, 1))
        configuration = {'x': 2}
        self.assertRaises(ValueError, Configuration, cs, configuration)

    def test_keys(self):
        # A regression test to make sure issue #49 does no longer pop up. By
        # iterating over the configuration in the for loop, it should not raise
        # a KeyError if the child hyperparameter is inactive.
        cs = ConfigurationSpace()
        shrinkage = CategoricalHyperparameter(
            "shrinkage", ["None", "auto", "manual"], default_value="None",
        )
        shrinkage_factor = UniformFloatHyperparameter(
            "shrinkage_factor", 0., 1., 0.5,
        )
        cs.add_hyperparameters([shrinkage, shrinkage_factor])

        cs.add_condition(EqualsCondition(shrinkage_factor, shrinkage, "manual"))

        for i in range(10):
            config = cs.sample_configuration()
            d = {hp_name: config[hp_name] for hp_name in
                 config if config[hp_name] is not None}



#!/usr/bin/env python

##
# wrapping: A program making it easy to use hyperparameter
# optimization software.
# Copyright (C) 2013 Katharina Eggensperger and Matthias Feurer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__authors__ = ["Katharina Eggensperger", "Matthias Feurer"]
__contact__ = "automl.org"

from collections import OrderedDict
from itertools import product
from io import StringIO
import sys
import json

import pyparsing

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
    UniformIntegerHyperparameter, UniformFloatHyperparameter, \
    NumericalHyperparameter, Constant, IntegerHyperparameter, \
    NormalIntegerHyperparameter, NormalFloatHyperparameter
from ConfigSpace.conditions import EqualsCondition, NotEqualsCondition,\
    InCondition, AndConjunction, OrConjunction, ConditionComponent
# from ConfigSpace.forbidden import ForbiddenEqualsClause, \
#     ForbiddenAndConjunction, ForbiddenInClause, AbstractForbiddenComponent, MultipleValueForbiddenClause
from ConfigSpace.forbidden import ForbiddenEqualsClause, \
    ForbiddenAndConjunction, ForbiddenInClause, AbstractForbiddenComponent, MultipleValueForbiddenClause


# Build pyparsing expressions for params
pp_param_name = pyparsing.Word(pyparsing.alphanums + "_" + "-" + "@" + "." + ":" + ";" + "\\" + "/" + "?" + "!" +
                               "$" + "%" + "&" + "*" + "+" + "<" + ">")
pp_digits = "0123456789"
pp_plusorminus = pyparsing.Literal('+') | pyparsing.Literal('-')
pp_int = pyparsing.Combine(pyparsing.Optional(pp_plusorminus) + pyparsing.Word(pp_digits))
pp_float = pyparsing.Combine(pyparsing.Optional(pp_plusorminus) + pyparsing.Optional(pp_int) + "." + pp_int)
pp_eorE = pyparsing.Literal('e') | pyparsing.Literal('E')
pp_floatorint = pp_float | pp_int
pp_e_notation = pyparsing.Combine(pp_floatorint + pp_eorE + pp_int)
pp_number = pp_e_notation | pp_float | pp_int
pp_numberorname = pp_number | pp_param_name
pp_il = pyparsing.Word("il")
pp_choices = pp_param_name + pyparsing.Optional(pyparsing.OneOrMore("," + pp_param_name))

pp_cont_param = pp_param_name + "[" + pp_number + "," + pp_number + "]" + \
    "[" + pp_number + "]" + pyparsing.Optional(pp_il)
pp_cat_param = pp_param_name + "{" + pp_choices + "}" + "[" + pp_param_name + "]"
pp_condition = pp_param_name + "|" + pp_param_name + "in" + "{" + pp_choices + "}"
pp_forbidden_clause = "{" + pp_param_name + "=" + pp_numberorname + \
    pyparsing.Optional(pyparsing.OneOrMore("," + pp_param_name + "=" + pp_numberorname)) + "}"


def build_categorical(param):
    cat_template = "%s {%s} [%s]"
    return cat_template % (param.name,
                           ", ".join([str(value) for value in param.choices]),
                           str(param.default_value))


def build_constant(param):
    constant_template = "%s {%s} [%s]"
    return constant_template % (param.name, param.value, param.value)


def build_continuous(param):
    if type(param) in (NormalIntegerHyperparameter,
                       NormalFloatHyperparameter):
        param = param.to_uniform()

    float_template = "%s%s [%s, %s] [%s]"
    int_template = "%s%s [%d, %d] [%d]i"
    if param.log:
        float_template += "l"
        int_template += "l"

    if param.q is not None:
        q_prefix = "Q%d_" % (int(param.q),)
    else:
        q_prefix = ""
    default_value = param.default_value

    if isinstance(param, IntegerHyperparameter):
        default_value = int(default_value)
        return int_template % (q_prefix, param.name, param.lower,
                               param.upper, default_value)
    else:
        return float_template % (q_prefix, param.name, str(param.lower),
                                 str(param.upper), str(default_value))


def build_condition(condition):
    if not isinstance(condition, ConditionComponent):
        raise TypeError("build_condition must be called with an instance of "
                        "'%s', got '%s'" %
                        (ConditionComponent, type(condition)))

    # Check if SMAC can handle the condition
    if isinstance(condition, OrConjunction):
        raise NotImplementedError("SMAC cannot handle OR conditions: %s" %
                                  (condition))
    if isinstance(condition, NotEqualsCondition):
        raise NotImplementedError("SMAC cannot handle != conditions: %s" %
                                 (condition))

    # Now handle the conditions SMAC can handle
    condition_template = "%s | %s in {%s}"
    if isinstance(condition, AndConjunction):
        return '\n'.join([
            build_condition(cond) for cond in condition.components
        ])
    elif isinstance(condition, InCondition):
        return condition_template % (condition.child.name,
                                     condition.parent.name,
                                     ", ".join(condition.values))
    elif isinstance(condition, EqualsCondition):
        return condition_template % (condition.child.name,
                                     condition.parent.name,
                                     condition.value)
    else:
        raise NotImplementedError(condition)


def build_forbidden(clause):
    if not isinstance(clause, AbstractForbiddenComponent):
        raise TypeError("build_forbidden must be called with an instance of "
                        "'%s', got '%s'" %
                        (AbstractForbiddenComponent, type(clause)))

    if not isinstance(clause, (ForbiddenEqualsClause, ForbiddenAndConjunction)):
        raise NotImplementedError("SMAC cannot handle '%s' of type %s" %
                                  str(clause), (type(clause)))

    retval = StringIO()
    retval.write("{")
    # Really simple because everything is an AND-conjunction of equals
    # conditions
    dlcs = clause.get_descendant_literal_clauses()
    for dlc in dlcs:
        if retval.tell() > 1:
            retval.write(", ")
        retval.write("%s=%s" % (dlc.hyperparameter.name, dlc.value))
    retval.write("}")
    retval.seek(0)
    return retval.getvalue()


def read(pcs_string, debug=False):
    configuration_space = ConfigurationSpace()
    conditions = []
    forbidden = []

    # some statistics
    ct = 0
    cont_ct = 0
    cat_ct = 0
    line_ct = 0

    for line in pcs_string:
        line_ct += 1

        if "#" in line:
            # It contains a comment
            pos = line.find("#")
            line = line[:pos]

        # Remove quotes and whitespaces at beginning and end
        ori_line = line
        line = line.replace('"', "").replace("'", "")
        line = line.strip()

        if "|" in line:
            # It's a condition
            try:
                c = pp_condition.parseString(line)
                conditions.append(c)
            except pyparsing.ParseException:
                raise NotImplementedError("Could not parse condition: %s" % line)
            continue
        if "}" not in line and "]" not in line:
            continue
        if line.startswith("{") and line.endswith("}"):
            forbidden.append(line)
            continue
        if len(line.strip()) == 0:
            continue

        ct += 1
        param = None

        create = {"int": UniformIntegerHyperparameter,
                  "float": UniformFloatHyperparameter,
                  "categorical": CategoricalHyperparameter}

        try:
            param_list = pp_cont_param.parseString(line)
            il = param_list[9:]
            if len(il) > 0:
                il = il[0]
            param_list = param_list[:9]
            name = param_list[0]
            lower = float(param_list[2])
            upper = float(param_list[4])
            paramtype = "int" if "i" in il else "float"
            log = True if "l" in il else False
            default_value = float(param_list[7])
            param = create[paramtype](name=name, lower=lower, upper=upper,
                                      q=None, log=log, default_value=default_value)
            cont_ct += 1
        except pyparsing.ParseException:
            pass

        try:
            param_list = pp_cat_param.parseString(line)
            name = param_list[0]
            #choices = [c for c in param_list[2:-4:2]]
            ori_line_segs = ori_line.split('{')
            choices_str = '[' + ori_line_segs[1].split('}')[0] + ']'
            choices = json.loads(choices_str)
            #default_value = param_list[-2]
            ori_line_segs = ori_line.split('[')
            default_str = ori_line_segs[-1].split(']')[0]
            default_value = json.loads(default_str)
            param = create["categorical"](name=name, choices=choices,
                                          default_value=default_value)
            cat_ct += 1
        except pyparsing.ParseException:
            pass

        if param is None:
            raise NotImplementedError("Could not parse: %s" % line)

        configuration_space.add_hyperparameter(param)

    for clause in forbidden:
        # TODO test this properly!
        # TODO Add a try/catch here!
        # noinspection PyUnusedLocal
        param_list = pp_forbidden_clause.parseString(clause)
        tmp_list = []
        clause_list = []
        for value in param_list[1:]:
            if len(tmp_list) < 3:
                tmp_list.append(value)
            else:
                # So far, only equals is supported by SMAC
                if tmp_list[1] == '=':
                    # TODO maybe add a check if the hyperparameter is
                    # actually in the configuration space
                    clause_list.append(ForbiddenEqualsClause(
                        configuration_space.get_hyperparameter(tmp_list[0]),
                        tmp_list[2]))
                else:
                    raise NotImplementedError()
                tmp_list = []
        configuration_space.add_forbidden_clause(ForbiddenAndConjunction(
            *clause_list))

    #Now handle conditions
    # If there are two conditions for one child, these two conditions are an
    # AND-conjunction of conditions, thus we have to connect them
    conditions_per_child = OrderedDict()
    for condition in conditions:
        child_name = condition[0]
        if child_name not in conditions_per_child:
            conditions_per_child[child_name] = list()
        conditions_per_child[child_name].append(condition)

    for child_name in conditions_per_child:
        condition_objects = []
        for condition in conditions_per_child[child_name]:
            child = configuration_space.get_hyperparameter(child_name)
            parent_name = condition[2]
            parent = configuration_space.get_hyperparameter(parent_name)
            restrictions = condition[5:-1:2]

            # TODO: cast the type of the restriction!
            if len(restrictions) == 1:
                condition = EqualsCondition(child, parent, restrictions[0])
            else:
                condition = InCondition(child, parent, values=restrictions)
            condition_objects.append(condition)

        # Now we have all condition objects for this child, so we can build a
        #  giant AND-conjunction of them (if number of conditions >= 2)!

        if len(condition_objects) > 1:
            and_conjunction = AndConjunction(*condition_objects)
            configuration_space.add_condition(and_conjunction)
        else:
            configuration_space.add_condition(condition_objects[0])

    return configuration_space


def write(configuration_space):
    if not isinstance(configuration_space, ConfigurationSpace):
        raise TypeError("pcs_parser.write expects an instance of %s, "
                        "you provided '%s'" % (ConfigurationSpace,
                        type(configuration_space)))

    param_lines = StringIO()
    condition_lines = StringIO()
    forbidden_lines = []
    for hyperparameter in configuration_space.get_hyperparameters():
        # Check if the hyperparameter names are valid SMAC names!
        try:
            pp_param_name.parseString(hyperparameter.name)
        except pyparsing.ParseException:
            raise ValueError(
                "Illegal hyperparameter name for SMAC: %s" % hyperparameter.name)

        # First build params
        if param_lines.tell() > 0:
            param_lines.write("\n")
        if isinstance(hyperparameter, NumericalHyperparameter):
            param_lines.write(build_continuous(hyperparameter))
        elif isinstance(hyperparameter, CategoricalHyperparameter):
            param_lines.write(build_categorical(hyperparameter))
        elif isinstance(hyperparameter, Constant):
            param_lines.write(build_constant(hyperparameter))
        else:
            raise TypeError("Unknown type: %s (%s)" % (
                type(hyperparameter), hyperparameter))

    for condition in configuration_space.get_conditions():
        if condition_lines.tell() > 0:
            condition_lines.write("\n")
        condition_lines.write(build_condition(condition))

    for forbidden_clause in configuration_space.get_forbiddens():
        # Convert in-statement into two or more equals statements
        dlcs = forbidden_clause.get_descendant_literal_clauses()
        # First, get all in statements and convert them to equal statements
        in_statements = []
        other_statements = []
        for dlc in dlcs:
            if isinstance(dlc, MultipleValueForbiddenClause):
                if not isinstance(dlc, ForbiddenInClause):
                    raise ValueError("SMAC cannot handle this forbidden "
                                     "clause: %s" % dlc)
                in_statements.append(
                    [ForbiddenEqualsClause(dlc.hyperparameter, value)
                     for value in dlc.values])
            else:
                other_statements.append(dlc)

        # Second, create the product of all elements in the IN statements,
        # create a ForbiddenAnd and add all ForbiddenEquals
        if len(in_statements) > 0:
            for i, p in enumerate(product(*in_statements)):
                all_forbidden_clauses = list(p) + other_statements
                f = ForbiddenAndConjunction(*all_forbidden_clauses)
                forbidden_lines.append(build_forbidden(f))
        else:
            forbidden_lines.append(build_forbidden(forbidden_clause))

    if condition_lines.tell() > 0:
        condition_lines.seek(0)
        param_lines.write("\n\n")
        for line in condition_lines:
            param_lines.write(line)

    if len(forbidden_lines) > 0:
        forbidden_lines.sort()
        param_lines.write("\n\n")
        for line in forbidden_lines:
            param_lines.write(line)
            param_lines.write("\n")

    # Check if the default configuration is a valid configuration!


    param_lines.seek(0)
    return param_lines.getvalue()


if __name__ == "__main__":
    fh = open(sys.argv[1])
    orig_pcs = fh.readlines()
    sp = read(orig_pcs, debug=True)
    created_pcs = write(sp).split("\n")
    print("============== Writing Results")
    print("#Lines: ", len(created_pcs))
    print("#LostLines: ", len(orig_pcs) - len(created_pcs))
    diff = ["%s\n" % i for i in created_pcs if i not in " ".join(orig_pcs)]
    print("Identical Lines: ", len(created_pcs) - len(diff))
    print()
    print("Up to 10 random different lines (of %d):" % len(diff))
    print("".join(diff[:10]))



# Copyright (c) 2014-2016, ConfigSpace developers
# Matthias Feurer
# Katharina Eggensperger
# and others (see commit history).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
import numpy as np
import warnings

from ConfigSpace.hyperparameters import Constant, \
    UniformFloatHyperparameter, NormalFloatHyperparameter, \
    UniformIntegerHyperparameter, NormalIntegerHyperparameter, \
    CategoricalHyperparameter, OrdinalHyperparameter
from ConfigSpace.conditions import EqualsCondition, NotEqualsCondition, \
    InCondition, AndConjunction, OrConjunction, LessThanCondition, \
    GreaterThanCondition


class TestConditions(unittest.TestCase):
    # TODO: return only copies of the objects!
    def test_equals_condition(self):
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cond = EqualsCondition(hp2, hp1, 0)
        cond_ = EqualsCondition(hp2, hp1, 0)

        # Test vector value:
        self.assertEqual(cond.vector_value, hp1._inverse_transform(0))
        self.assertEqual(cond.vector_value, cond_.vector_value)

        # Test invalid conditions:
        self.assertRaises(TypeError, EqualsCondition, hp2, "parent", 0)
        self.assertRaises(TypeError, EqualsCondition, "child", hp1, 0)
        self.assertRaises(ValueError, EqualsCondition, hp1, hp1, 0)

        self.assertEqual(cond, cond_)

        cond_reverse = EqualsCondition(hp1, hp2, 0)
        self.assertNotEqual(cond, cond_reverse)

        self.assertNotEqual(cond, dict())

        self.assertEqual("child | parent == 0", str(cond))

    def test_equals_condition_illegal_value(self):
        epsilon = UniformFloatHyperparameter("epsilon", 1e-5, 1e-1,
                                             default_value=1e-4, log=True)
        loss = CategoricalHyperparameter("loss",
                                         ["hinge", "log", "modified_huber", "squared_hinge", "perceptron"],
                                         default_value="hinge")
        self.assertRaisesRegexp(ValueError, "Hyperparameter 'epsilon' is "
                                            "conditional on the illegal value 'huber' of "
                                            "its parent hyperparameter 'loss'",
                                EqualsCondition, epsilon, loss, "huber")

    def test_not_equals_condition(self):
        hp1 = CategoricalHyperparameter("parent", [0, 1])
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cond = NotEqualsCondition(hp2, hp1, 0)
        cond_ = NotEqualsCondition(hp2, hp1, 0)
        self.assertEqual(cond, cond_)

        # Test vector value:
        self.assertEqual(cond.vector_value, hp1._inverse_transform(0))
        self.assertEqual(cond.vector_value, cond_.vector_value)

        cond_reverse = NotEqualsCondition(hp1, hp2, 0)
        self.assertNotEqual(cond, cond_reverse)

        self.assertNotEqual(cond, dict())

        self.assertEqual("child | parent != 0", str(cond))

    def test_not_equals_condition_illegal_value(self):
        epsilon = UniformFloatHyperparameter("epsilon", 1e-5, 1e-1,
                                             default_value=1e-4, log=True)
        loss = CategoricalHyperparameter("loss",
                                         ["hinge", "log", "modified_huber",
                                          "squared_hinge", "perceptron"],
                                         default_value="hinge")
        self.assertRaisesRegexp(ValueError, "Hyperparameter 'epsilon' is "
                                            "conditional on the illegal value 'huber' of "
                                            "its parent hyperparameter 'loss'",
                                NotEqualsCondition, epsilon, loss, "huber")

    def test_in_condition(self):
        hp1 = CategoricalHyperparameter("parent", list(range(0, 11)))
        hp2 = UniformIntegerHyperparameter("child", 0, 10)
        cond = InCondition(hp2, hp1, [0, 1, 2, 3, 4, 5])
        cond_ = InCondition(hp2, hp1, [0, 1, 2, 3, 4, 5])
        self.assertEqual(cond, cond_)

        # Test vector value:
        self.assertEqual(cond.vector_values, [hp1._inverse_transform(i) for i in [0, 1, 2, 3, 4, 5]])
        self.assertEqual(cond.vector_values, cond_.vector_values)

        cond_reverse = InCondition(hp1, hp2, [0, 1, 2, 3, 4, 5])
        self.assertNotEqual(cond, cond_reverse)

        self.assertNotEqual(cond, dict())

        self.assertEqual("child | parent in {0, 1, 2, 3, 4, 5}", str(cond))

    def test_greater_and_less_condition(self):
        child = Constant('child', 'child')
        hp1 = UniformFloatHyperparameter("float", 0, 5)
        hp2 = UniformIntegerHyperparameter("int", 0, 5)
        hp3 = OrdinalHyperparameter("ord", list(range(6)))

        for hp in [hp1, hp2, hp3]:
            hyperparameter_idx = {
                child.name: 0,
                hp.name: 1
            }

            gt = GreaterThanCondition(child, hp, 1)
            gt.set_vector_idx(hyperparameter_idx)
            self.assertFalse(gt.evaluate({hp.name: 0}))
            self.assertTrue(gt.evaluate({hp.name: 2}))
            self.assertFalse(gt.evaluate({hp.name: None}))

            # Evaluate vector
            test_value = hp._inverse_transform(2)
            self.assertFalse(gt.evaluate_vector(np.array([np.NaN, 0])))
            self.assertTrue(gt.evaluate_vector(np.array([np.NaN, test_value])))
            self.assertFalse(gt.evaluate_vector(np.array([np.NaN, np.NaN])))

            lt = LessThanCondition(child, hp, 1)
            lt.set_vector_idx(hyperparameter_idx)
            self.assertTrue(lt.evaluate({hp.name: 0}))
            self.assertFalse(lt.evaluate({hp.name: 2}))
            self.assertFalse(lt.evaluate({hp.name: None}))

            # Evaluate vector
            test_value = hp._inverse_transform(2)
            self.assertTrue(lt.evaluate_vector(np.array([np.NaN, 0, 0, 0])))
            self.assertFalse(lt.evaluate_vector(np.array([np.NaN, test_value])))
            self.assertFalse(lt.evaluate_vector(np.array([np.NaN, np.NaN])))

        hp4 = CategoricalHyperparameter("cat", list(range(6)))
        self.assertRaisesRegexp(ValueError, "Parent hyperparameter in a > or < "
                                            "condition must be a subclass of "
                                            "NumericalHyperparameter or "
                                            "OrdinalHyperparameter, but is "
                                            "<cdef class 'ConfigSpace.hyperparameters.CategoricalHyperparameter'>",
                                GreaterThanCondition, child, hp4, 1)
        self.assertRaisesRegexp(ValueError, "Parent hyperparameter in a > or < "
                                            "condition must be a subclass of "
                                            "NumericalHyperparameter or "
                                            "OrdinalHyperparameter, but is "
                                            "<cdef class 'ConfigSpace.hyperparameters.CategoricalHyperparameter'>",
                                LessThanCondition, child, hp4, 1)

        hp5 = OrdinalHyperparameter("ord", ['cold', 'luke warm', 'warm', 'hot'])

        hyperparameter_idx = {
            child.name: 0,
            hp5.name: 1
        }
        gt = GreaterThanCondition(child, hp5, 'warm')
        gt.set_vector_idx(hyperparameter_idx)
        self.assertTrue(gt.evaluate({hp5.name: 'hot'}))
        self.assertFalse(gt.evaluate({hp5.name: 'cold'}))

        self.assertTrue(gt.evaluate_vector(np.array([np.NaN, 3])))
        self.assertFalse(gt.evaluate_vector(np.array([np.NaN, 0])))

        lt = LessThanCondition(child, hp5, 'warm')
        lt.set_vector_idx(hyperparameter_idx)
        self.assertTrue(lt.evaluate({hp5.name: 'luke warm'}))
        self.assertFalse(lt.evaluate({hp5.name: 'warm'}))

        self.assertTrue(lt.evaluate_vector(np.array([np.NaN, 1])))
        self.assertFalse(lt.evaluate_vector(np.array([np.NaN, 2])))

    def test_in_condition_illegal_value(self):
        epsilon = UniformFloatHyperparameter("epsilon", 1e-5, 1e-1,
                                             default_value=1e-4, log=True)
        loss = CategoricalHyperparameter("loss",
                                         ["hinge", "log", "modified_huber",
                                          "squared_hinge", "perceptron"],
                                         default_value="hinge")
        self.assertRaisesRegexp(ValueError, "Hyperparameter 'epsilon' is "
                                            "conditional on the illegal value 'huber' of "
                                            "its parent hyperparameter 'loss'",
                                InCondition, epsilon, loss, ["huber", "log"])

    def test_and_conjunction(self):
        self.assertRaises(TypeError, AndConjunction, "String1", "String2")

        hp1 = CategoricalHyperparameter("input1", [0, 1])
        hp2 = CategoricalHyperparameter("input2", [0, 1])
        hp3 = CategoricalHyperparameter("input3", [0, 1])
        hp4 = Constant("And", "True")
        cond1 = EqualsCondition(hp4, hp1, 1)

        # Only one condition in an AndConjunction!
        self.assertRaises(ValueError, AndConjunction, cond1)

        cond2 = EqualsCondition(hp4, hp2, 1)
        cond3 = EqualsCondition(hp4, hp3, 1)

        andconj1 = AndConjunction(cond1, cond2)
        andconj1_ = AndConjunction(cond1, cond2)
        self.assertEqual(andconj1, andconj1_)

        # Test setting vector idx
        hyperparameter_idx = {
            hp1.name: 0,
            hp2.name: 1,
            hp3.name: 2,
            hp4.name: 3
        }
        andconj1.set_vector_idx(hyperparameter_idx)
        self.assertEqual(andconj1.get_parents_vector(), [0, 1])
        self.assertEqual(andconj1.get_children_vector(), [3, 3])

        andconj2 = AndConjunction(cond2, cond3)
        self.assertNotEqual(andconj1, andconj2)

        andconj3 = AndConjunction(cond1, cond2, cond3)
        self.assertEqual("(And | input1 == 1 && And | input2 == 1 && And | "
                         "input3 == 1)", str(andconj3))

        # Test __eq__
        self.assertNotEqual(andconj1, andconj3)
        self.assertNotEqual(andconj1, "String")

    def test_or_conjunction(self):
        self.assertRaises(TypeError, AndConjunction, "String1", "String2")

        hp1 = CategoricalHyperparameter("input1", [0, 1])
        hp2 = CategoricalHyperparameter("input2", [0, 1])
        hp3 = CategoricalHyperparameter("input3", [0, 1])
        hp4 = Constant("Or", "True")
        cond1 = EqualsCondition(hp4, hp1, 1)

        self.assertRaises(ValueError, OrConjunction, cond1)

        cond2 = EqualsCondition(hp4, hp2, 1)
        cond3 = EqualsCondition(hp4, hp3, 1)

        andconj1 = OrConjunction(cond1, cond2)
        andconj1_ = OrConjunction(cond1, cond2)
        self.assertEqual(andconj1, andconj1_)

        # Test setting vector idx
        hyperparameter_idx = {
            hp1.name: 0,
            hp2.name: 1,
            hp3.name: 2,
            hp4.name: 3
        }
        andconj1.set_vector_idx(hyperparameter_idx)
        self.assertEqual(andconj1.get_parents_vector(), [0, 1])
        self.assertEqual(andconj1.get_children_vector(), [3, 3])

        andconj2 = OrConjunction(cond2, cond3)
        self.assertNotEqual(andconj1, andconj2)

        andconj3 = OrConjunction(cond1, cond2, cond3)
        self.assertEqual("(Or | input1 == 1 || Or | input2 == 1 || Or | "
                         "input3 == 1)", str(andconj3))

    def test_nested_conjunctions(self):
        hp1 = CategoricalHyperparameter("input1", [0, 1])
        hp2 = CategoricalHyperparameter("input2", [0, 1])
        hp3 = CategoricalHyperparameter("input3", [0, 1])
        hp4 = CategoricalHyperparameter("input4", [0, 1])
        hp5 = CategoricalHyperparameter("input5", [0, 1])
        hp6 = Constant("AND", "True")

        cond1 = EqualsCondition(hp6, hp1, 1)
        cond2 = EqualsCondition(hp6, hp2, 1)
        cond3 = EqualsCondition(hp6, hp3, 1)
        cond4 = EqualsCondition(hp6, hp4, 1)
        cond5 = EqualsCondition(hp6, hp5, 1)

        conj1 = AndConjunction(cond1, cond2)
        conj2 = OrConjunction(conj1, cond3)
        conj3 = AndConjunction(conj2, cond4, cond5)

        # TODO: this does not look nice, And should depend on a large
        # conjunction, there should not be many ANDs inside this string!
        self.assertEqual("(((AND | input1 == 1 && AND | input2 == 1) || AND | "
                         "input3 == 1) && AND | input4 == 1 && AND | input5 "
                         "== 1)", str(conj3))

    def test_all_components_have_the_same_child(self):
        hp1 = CategoricalHyperparameter("input1", [0, 1])
        hp2 = CategoricalHyperparameter("input2", [0, 1])
        hp3 = CategoricalHyperparameter("input3", [0, 1])
        hp4 = CategoricalHyperparameter("input4", [0, 1])
        hp5 = CategoricalHyperparameter("input5", [0, 1])
        hp6 = Constant("AND", "True")

        cond1 = EqualsCondition(hp1, hp2, 1)
        cond2 = EqualsCondition(hp1, hp3, 1)
        cond3 = EqualsCondition(hp1, hp4, 1)
        cond4 = EqualsCondition(hp6, hp4, 1)
        cond5 = EqualsCondition(hp6, hp5, 1)

        AndConjunction(cond1, cond2, cond3)
        AndConjunction(cond4, cond5)
        self.assertRaisesRegexp(ValueError,
                                "All Conjunctions and Conditions must have "
                                "the same child.", AndConjunction, cond1, cond4)

    def test_condition_from_cryptominisat(self):
        parent = CategoricalHyperparameter('blkrest', ['0', '1'], default_value='1')
        child = UniformIntegerHyperparameter('blkrestlen', 2000, 10000,
                                             log=True)
        condition = EqualsCondition(child, parent, '1')
        self.assertFalse(condition.evaluate(dict(blkrest='0')))
        self.assertTrue(condition.evaluate(dict(blkrest='1')))

    def test_get_parents(self):
        # Necessary because we couldn't call cs.get_parents for
        # clasp-sat-params-nat.pcs
        counter = UniformIntegerHyperparameter('bump', 10, 4096, log=True)
        _1_S_countercond = CategoricalHyperparameter('cony', ['yes', 'no'])
        _1_0_restarts = CategoricalHyperparameter('restarts', ['F', 'L', 'D',
                                                               'x', '+', 'no'],
                                                  default_value='x')

        condition = EqualsCondition(counter, _1_S_countercond, 'yes')
        # All conditions inherit get_parents from abstractcondition
        self.assertEqual([_1_S_countercond], condition.get_parents())
        condition2 = InCondition(counter, _1_0_restarts, ['F', 'D', 'L', 'x', '+'])
        # All conjunctions inherit get_parents from abstractconjunction
        conjunction = AndConjunction(condition, condition2)
        self.assertEqual([_1_S_countercond, _1_0_restarts], conjunction.get_parents())



# Copyright (c) 2014-2016, ConfigSpace developers
# Matthias Feurer
# Katharina Eggensperger
# and others (see commit history).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# import unittest
#
# from hyperopt import hp
# import hyperopt
# import numpy as np
# import six
#
# import HPOlib.format_converter.configuration_space as configuration_space
# import HPOlib.format_converter.pyll_parser as pyll_parser
# from HPOlibConfigSpace.hyperparameters import CategoricalHyperparameter, \
#     UniformFloatHyperparameter, UniformIntegerHyperparameter, Constant, \
#     NormalFloatHyperparameter, NormalIntegerHyperparameter
#
# # More complex search space
# classifier = CategoricalHyperparameter("classifier", ["svm", "nn"])
# kernel = CategoricalHyperparameter("kernel", ["rbf", "linear"],
#     conditions=[["classifier == svm"]])
# C = UniformFloatHyperparameter("C", 0.03125, 32768, base=2,
#     conditions=[["classifier == svm"]])
# gamma = UniformFloatHyperparameter("gamma", 0.000030518, 8, base=2,
#     conditions=[["kernel == rbf"]])
# neurons = UniformIntegerHyperparameter("neurons", 16, 1024, q=16,
#     conditions=[["classifier == nn"]])
# lr = UniformFloatHyperparameter("lr", 0.0001, 1.0,
#     conditions=[["classifier == nn"]])
# preprocessing = CategoricalHyperparameter("preprocessing", [None, "pca"])
# config_space = {"classifier": classifier,
#                 "kernel": kernel,
#                 "C": C,
#                 "gamma": gamma,
#                 "neurons": neurons,
#                 "lr": lr,
#                 "preprocessing": preprocessing}
#
# # A search space where a hyperparameter depends on two others:
# gamma_2 = UniformFloatHyperparameter("gamma_2", 0.000030518, 8, base=2,
#     conditions=[["kernel == rbf", "classifier == svm"]])
#
# config_space_2 = {"classifier": classifier,
#                 "kernel": kernel,
#                 "C": C,
#                 "gamma_2": gamma_2,
#                 "neurons": neurons,
#                 "lr": lr,
#                 "preprocessing": preprocessing}
#
#
# class TestPyllReader(unittest.TestCase):
#     def setUp(self):
#         self.pyll_reader = pyll_parser.PyllReader()
#
#     def test_read_literal(self):
#         literal = hyperopt.pyll.as_apply("5....4....3....1! Off blast! ")
#         ret = self.pyll_reader.read_literal(literal, "pre_chorus")
#         expected = Constant("pre_chorus", "5....4....3....1! Off blast! ")
#         self.assertEqual(expected, ret)
#
#     def test_read_container(self):
#         #### Lists
#         # The Literal is added to the content of the list, but no method will
#         #  add it to the list of found hyperparameters
#         # hyperparameter
#         # 0 pos_args
#         # 1   float
#         # 2     hyperopt_param
#         # 3       Literal{a}
#         # 4       uniform
#         # 5         Literal{0}
#         # 6         Literal{10}
#         # 7   Literal{Alpha}
#         expected = {'a': UniformFloatHyperparameter('a', 0, 10),
#                     'Alpha': Constant('Alpha', 'Alpha')}
#         param = hyperopt.pyll.as_apply([hp.uniform("a", 0, 10), "Alpha"])
#         ret = self.pyll_reader.read_container(param)
#         self.assertEqual(expected, ret)
#         # 0 pos_args
#         # 1   float
#         # 2     hyperopt_param
#         # 3       Literal{a}
#         # 4       uniform
#         # 5         Literal{0}
#         # 6         Literal{10}
#         # 7   float
#         # 8     hyperopt_param
#         # 9       Literal{b}
#         # 10       uniform
#         # 11         Literal{5}
#         # 12         Literal{15}
#         container =  hyperopt.pyll.as_apply([hp.uniform("a", 0, 10),
#                                              hp.uniform("b", 5, 15)])
#         ret = self.pyll_reader.read_container(container)
#         expected = {"a": UniformFloatHyperparameter("a", 0, 10),
#                     "b": UniformFloatHyperparameter("b", 5, 15)}
#         self.assertEqual(expected, ret)
#
#     def test_read_dict(self):
#         #### Dictionaries
#         # 0 dict
#         # 1  Elektronik =
#         # 2   Literal{Supersonik}
#         # 3  a =
#         # 4   float
#         # 5     hyperopt_param
#         # 6       Literal{a}
#         # 7       uniform
#         # 8         Literal{0}
#         # 9         Literal{10}
#         container = hyperopt.pyll.as_apply({"a": hp.uniform("a", 0, 10),
#                                             "Elektronik": "Supersonik"})
#         ret = self.pyll_reader.read_dict(container)
#         expected = {"a": UniformFloatHyperparameter("a", 0, 10),
#                     "Elektronik": Constant("Elektronik", "Supersonik")}
#         self.assertEqual(expected, ret)
#
#         # 0 dict
#         # 1  @1:F:ASPEED:aspeed-opt =
#         # 2   switch
#         # 3     hyperopt_param
#         # 4       Literal{@1:F:ASPEED:aspeed-opt}
#         # 5       randint
#         # 6         Literal{1}
#         # 7     dict
#         # 8      @1:F:ASPEED:aspeed-opt =
#         # 9       Literal{yes}
#         # 10  @1:approach =
#         # 11   Literal{ASPEED}
#         param_6 = hp.choice("@1:F:ASPEED:aspeed-opt", [
#                                 {"@1:F:ASPEED:aspeed-opt": "yes", },
#                                 ])
#         container = hyperopt.pyll.as_apply(
#             {"@1:F:ASPEED:aspeed-opt": param_6,
#              "@1:approach": "ASPEED"})
#         ret = self.pyll_reader.read_dict(container)
#         expected = {"@1:F:ASPEED:aspeed-opt":
#                         CategoricalHyperparameter("@1:F:ASPEED:aspeed-opt", ['yes']),
#                     "@1:approach":
#                         Constant("@1:approach", "ASPEED")}
#         self.maxDiff = None
#         self.assertEqual(expected, ret)
#
#     def test_read_switch(self):
#         # 0 switch
#         # 1   hyperopt_param
#         # 2     Literal{dist1}
#         # 3     randint
#         # 4       Literal{2}
#         # 5   Literal{uniform}
#         # 6   Literal{normal}
#         dist = hp.choice('dist1', ['uniform', 'normal'])
#         ret = self.pyll_reader.read_switch(dist)
#         expected = CategoricalHyperparameter('dist1', ['uniform', 'normal'])
#         self.assertEqual(expected, ret)
#
#         bigger_choice = hp.choice('choice', [
#             {'choice': "zero", 'a': 0, 'b': hp.uniform('b', 0, 10)},
#             {'choice': "other", 'a': 1, 'b': hp.uniform('b', 0, 10)}])
#         ret = self.pyll_reader.read_switch(bigger_choice)
#         expected = CategoricalHyperparameter('choice', ['zero', 'other'])
#         self.assertEqual(expected, ret)
#         self.assertEqual(2, len(self.pyll_reader.constants))
#         # Only the hyperparameter b is put into pyll_reader.hyperparameters
#         self.assertEqual(1, len(self.pyll_reader.hyperparameters))
#
#     # TODO: duplicate these tests for Integer/care about integers + test if
#     # the warning of non-uniform parameters is actually printed
#     def test_read_uniform(self):
#         # 0 float
#         # 1   hyperopt_param
#         # 2     Literal{scale_mult1}
#         # 3     uniform
#         # 4       Literal{0.2}
#         # 5       Literal{2}
#         uniform = hp.uniform('scale_mult1', .2, 2).inputs()[0].inputs()[1]
#         ret = self.pyll_reader.read_uniform(uniform, 'scale_mult1')
#         expected = UniformFloatHyperparameter('scale_mult1', 0.2, 2)
#         self.assertEqual(expected, ret)
#
#     def test_read_loguniform(self):
#         # 0 float
#         # 1   hyperopt_param
#         # 2     Literal{colnorm_thresh}
#         # 3     loguniform
#         # 4       Literal{-20.7232658369}
#         # 5       Literal{-6.90775527898}
#         loguniform = hp.loguniform('colnorm_thresh', np.log(1e-9),
#             np.log(1e-3)).inputs()[0].inputs()[1]
#         ret = self.pyll_reader.read_loguniform(loguniform, 'colnorm_thresh')
#         expected = UniformFloatHyperparameter(
#             'colnorm_thresh', 1e-9, 1e-3, base=np.e)
#         self.assertEqual(expected, ret)
#
#     def test_read_quniform(self):
#         # TODO scope.int(hp.quniform('liblinear:LOG2_C', -5, 15, 1))
#         # 0 float
#         # 1   hyperopt_param
#         # 2     Literal{l0eg_fsize}
#         # 3     quniform
#         # 4       Literal{2.50001}
#         # 5       Literal{8.5}
#         # 6       Literal{1}
#         quniform = hp.quniform('l0eg_fsize', 2.50001, 8.5, 1). \
#             inputs()[0].inputs()[1]
#         ret = self.pyll_reader.read_quniform(quniform, 'l0eg_fsize')
#         expected = UniformIntegerHyperparameter(
#             'l0eg_fsize', 3, 8)
#         self.assertEqual(expected, ret)
#
#         l2_out_lp_psize = hp.quniform("l2_out_lp_psize", 0.50001, 5.5, 1). \
#             inputs()[0].inputs()[1]
#         ret = self.pyll_reader.read_quniform(l2_out_lp_psize, "l2_out_lp_psize")
#         expected = UniformIntegerHyperparameter(
#             "l2_out_lp_psize", 1, 5)
#         self.assertEqual(expected, ret)
#
#     def test_read_qloguniform(self):
#         # 0 float
#         # 1   hyperopt_param
#         # 2     Literal{nhid1}
#         # 3     qloguniform
#         # 4       Literal{2.77258872224}
#         # 5       Literal{6.9314718056}
#         # 6      q =
#         # 7       Literal{16}
#         qloguniform = hp.qloguniform('nhid1', np.log(16), np.log(1024), q=16). \
#             inputs()[0].inputs()[1]
#         ret = self.pyll_reader.read_qloguniform(qloguniform, 'nhid1')
#         expected = UniformFloatHyperparameter(
#             'nhid1', 16, 1024, q=16, base=np.e)
#         self.assertEqual(expected, ret)
#
#     def test_read_normal(self):
#         # 0 float
#         # 1   hyperopt_param
#         # 2     Literal{l0eg_alpha}
#         # 3     normal
#         # 4       Literal{0.0}
#         # 5       Literal{1.0}
#         normal = hp.normal("l0eg_alpha", 0.0, 1.0).inputs()[0].inputs()[1]
#         ret = self.pyll_reader.read_normal(normal, "l0eg_alpha")
#         expected = NormalFloatHyperparameter("l0eg_alpha", 0.0, 1.0)
#         self.assertEqual(expected, ret)
#
#
#     def test_read_lognormal(self):
#         # 0 float
#         # 1   hyperopt_param
#         # 2     Literal{lr}
#         # 3     lognormal
#         # 4       Literal{-4.60517018599}
#         # 5       Literal{3.0}
#         lognormal = hp.lognormal('lr', np.log(.01), 3.).inputs()[0].inputs()[1]
#         ret = self.pyll_reader.read_lognormal(lognormal, "lr")
#         expected = NormalFloatHyperparameter("lr", np.log(0.01), 3.0, base=np.e)
#         self.assertEqual(expected, ret)
#
#     def test_read_qnormal(self):
#         # 0 float
#         # 1   hyperopt_param
#         # 2     Literal{qnormal}
#         # 3     qnormal
#         # 4       Literal{0.0}
#         # 5       Literal{1.0}
#         # 6       Literal{0.5}
#         qnormal = hp.qnormal("qnormal", 0.0, 1.0, 0.5).inputs()[0].inputs()[1]
#         ret = self.pyll_reader.read_qnormal(qnormal, "qnormal")
#         expected = NormalFloatHyperparameter("qnormal", 0.0, 1.0, q=0.5)
#         self.assertEqual(expected, ret)
#
#     def test_read_qlognormal(self):
#         # 0 float
#         # 1   hyperopt_param
#         # 2     Literal{qlognormal}
#         # 3     qlognormal
#         # 4       Literal{0.0}
#         # 5       Literal{1.0}
#         # 6       Literal{0.5}
#         qlognormal = hp.qlognormal("qlognormal", 0.0, 1.0, 0.5).inputs()[0].inputs()[1]
#         ret = self.pyll_reader.read_qlognormal(qlognormal, "qlognormal")
#         expected = NormalFloatHyperparameter("qlognormal", 0.0, 1.0, q=0.5, base=np.e)
#         self.assertEqual(expected, ret)
#
#         qlognormal = hp.qlognormal("qlognormal", 1.0, 5.0, 1.0).inputs()[0].inputs()[1]
#         ret = self.pyll_reader.read_qlognormal(qlognormal, "qlognormal")
#         expected = NormalIntegerHyperparameter("qlognormal", 1.0, 5.0, base=np.e)
#         self.assertEqual(expected, ret)
#
#
# class TestPyllWriter(unittest.TestCase):
#     def setUp(self):
#         self.pyll_writer = pyll_parser.PyllWriter()
#
#     def test_convert_configuration_space(self):
#         a = UniformFloatHyperparameter("a", 0, 1)
#         b = UniformFloatHyperparameter("b", 0, 3, q=0.1)
#
#         expected = six.StringIO()
#         expected.write('from hyperopt import hp\nimport hyperopt.pyll as pyll')
#         expected.write('\n\n')
#         expected.write('param_0 = hp.uniform("a", 0.0, 1.0)\n')
#         expected.write('param_1 = hp.quniform("b", -0.0499, 3.05, 0.1)\n\n')
#         expected.write('space = {"a": param_0, "b": param_1}\n')
#         simple_space = {"a": a, "b": b}
#         cs = self.pyll_writer.write(simple_space)
#         self.assertEqual(expected.getvalue(), cs)
#
#     def test_convert_conditional_space(self):
#         a_or_b = CategoricalHyperparameter("a_or_b", ["a", "b"])
#         cond_a = UniformFloatHyperparameter(
#             'cond_a', 0, 1, conditions=[['a_or_b == a']])
#         cond_b = UniformFloatHyperparameter(
#             'cond_b', 0, 3, q=0.1, conditions=[['a_or_b == b']])
#         conditional_space = {"a_or_b": a_or_b, "cond_a": cond_a, "cond_b": cond_b}
#         cs = self.pyll_writer.write(conditional_space)
#         expected = six.StringIO()
#         expected.write('from hyperopt import hp\nimport hyperopt.pyll as pyll')
#         expected.write('\n\n')
#         expected.write('param_0 = hp.uniform("cond_a", 0.0, 1.0)\n')
#         expected.write('param_1 = hp.quniform("cond_b", -0.0499, 3.05, 0.1)\n')
#         expected.write('param_2 = hp.choice("a_or_b", [\n')
#         expected.write('    {"a_or_b": "a", "cond_a": param_0, },\n')
#         expected.write('    {"a_or_b": "b", "cond_b": param_1, },\n')
#         expected.write('    ])\n\n')
#         expected.write('space = {"a_or_b": param_2}\n')
#         self.assertEqual(expected.getvalue(), cs)
#
#     def test_convert_complex_space(self):
#         cs = self.pyll_writer.write(config_space)
#         expected = six.StringIO()
#         expected.write('from hyperopt import hp\nimport hyperopt.pyll as pyll')
#         expected.write('\n\n')
#         expected.write('param_0 = hp.uniform("LOG2_C", -5.0, 15.0)\n')
#         expected.write('param_1 = hp.uniform("LOG2_gamma", -14.9999800563, '
#                        '3.0)\n')
#         expected.write('param_2 = hp.choice("kernel", [\n')
#         expected.write('    {"kernel": "linear", },\n')
#         expected.write('    {"kernel": "rbf", "LOG2_gamma": param_1, },\n')
#         expected.write('    ])\n')
#         expected.write('param_3 = hp.uniform("lr", 0.0001, 1.0)\n')
#         expected.write('param_4 = pyll.scope.int(hp.quniform('
#                        '"neurons", 15.50001, 1024.5, 16.0))\n')
#         expected.write('param_5 = hp.choice("classifier", [\n')
#         expected.write('    {"classifier": "nn", "lr": param_3, "neurons": '
#                        'param_4, },\n')
#         expected.write('    {"classifier": "svm", "LOG2_C": param_0, '
#                        '"kernel": param_2, },\n')
#         expected.write('    ])\n')
#         expected.write('param_6 = hp.choice("preprocessing", [\n')
#         expected.write('    {"preprocessing": "None", },\n')
#         expected.write('    {"preprocessing": "pca", },\n')
#         expected.write('    ])\n\n')
#         expected.write('space = {"classifier": param_5, '
#                        '"preprocessing": param_6}\n')
#         self.assertEqual(expected.getvalue(), cs)
#
#         self.pyll_writer.reset_hyperparameter_countr()
#         expected.seek(0)
#         cs = self.pyll_writer.write(config_space_2)
#         self.assertEqual(expected.getvalue().replace("gamma", "gamma_2"), cs)
#
#     def test_operator_in(self):
#         a_or_b = CategoricalHyperparameter("a_or_b", ["a", "b"])
#         cond_a = UniformFloatHyperparameter(
#             'cond_a', 0, 1, conditions=[['a_or_b == a']])
#         cond_b = UniformFloatHyperparameter(
#             'cond_b', 0, 3, q=0.1, conditions=[['a_or_b == b']])
#         e = UniformFloatHyperparameter("e", 0, 5,
#                                      conditions=[['a_or_b in {a,b}']])
#         conditional_space_operator_in = {"a_or_b": a_or_b, "cond_a": cond_a,
#                                  "cond_b": cond_b, "e": e}
#         cs = self.pyll_writer.write(conditional_space_operator_in)
#         expected = six.StringIO()
#         expected.write('from hyperopt import hp\nimport hyperopt.pyll as pyll')
#         expected.write('\n\n')
#         expected.write('param_0 = hp.uniform("cond_a", 0.0, 1.0)\n')
#         expected.write('param_1 = hp.quniform("cond_b", -0.0499, 3.05, 0.1)\n')
#         expected.write('param_2 = hp.uniform("e", 0.0, 5.0)\n')
#         expected.write('param_3 = hp.choice("a_or_b", [\n')
#         expected.write('    {"a_or_b": "a", "cond_a": param_0, "e": param_2, '
#                        '},\n')
#         expected.write('    {"a_or_b": "b", "cond_b": param_1, "e": param_2, '
#                        '},\n')
#         expected.write('    ])\n\n')
#         expected.write('space = {"a_or_b": param_3}\n')
#         self.assertEqual(expected.getvalue(), cs)
#
#     def test_write_uniform(self):
#         a = UniformFloatHyperparameter("a", 0, 1)
#         expected = ('a', 'param_0 = hp.uniform("a", 0.0, 1.0)')
#         value = self.pyll_writer.write_hyperparameter(a, None)
#         self.assertEqual(expected, value)
#
#         # The hyperparameter name has to be converted seperately because
#         # otherwise the parameter values are converted at object costruction
#         # time
#         a = UniformFloatHyperparameter("a", 1, 10, base=10)
#         a.name = self.pyll_writer.convert_name(a)
#         expected = ('LOG10_a', 'param_1 = hp.uniform("LOG10_a", 0.0, 1.0)')
#         value = self.pyll_writer.write_hyperparameter(a, None)
#         self.assertEqual(expected, value)
#
#         nhid1 = UniformFloatHyperparameter(
#             "nhid1", 16, 1024, q=16, base=np.e)
#         expected = ('nhid1', 'param_2 = hp.qloguniform('
#                     '"nhid1", 2.0794540416, 6.93925394604, 16.0)')
#         value = self.pyll_writer.write_hyperparameter(nhid1, None)
#         self.assertEqual(expected, value)
#
#     def test_write_uniform_int(self):
#         a_int = UniformIntegerHyperparameter("a_int", 0, 1)
#         expected = ('a_int', 'param_0 = pyll.scope.int(hp.quniform('
#                              '"a_int", -0.49999, 1.5, 1.0))')
#         value = self.pyll_writer.write_hyperparameter(a_int, None)
#         self.assertEqual(expected, value)
#
#         # Test for the problem that if a parameter has Q not None and is on
#         # log scale, the Q must not be in the hp object, but the
#         # hyperparameter name. If this is done the other way round,
#         # the log-value of the hyperparameter is quantized
#         a_int = UniformIntegerHyperparameter(
#             "a_int", 1, 1000, base=10)
#         a_int.name = self.pyll_writer.convert_name(a_int)
#         expected = ('LOG10_Q1_a_int', 'param_1 = hp.uniform('
#                     '"LOG10_Q1_a_int", -0.301021309861, 3.00021709297)')
#         value = self.pyll_writer.write_hyperparameter(a_int, None)
#         self.assertEqual(expected, value)
#
#     def test_write_quniform(self):
#         b = UniformFloatHyperparameter("b", 0, 3, q=0.1)
#         expected = ("b", 'param_0 = hp.quniform("b", -0.0499, 3.05, 0.1)')
#         value = self.pyll_writer.write_hyperparameter(b, None)
#         self.assertEqual(expected, value)
#
#         b = UniformFloatHyperparameter(
#             "b", 0.1, 3, q=0.1, base=10)
#         b.name = self.pyll_writer.convert_name(b)
#         expected = ('LOG10_Q0.100000_b', 'param_1 = hp.uniform('
#                     '"LOG10_Q0.100000_b", -1.30016227413, 0.484299839347)')
#         value = self.pyll_writer.write_hyperparameter(b, None)
#         self.assertEqual(expected, value)
#
#     def test_write_quniform_int(self):
#         b_int_1 = UniformIntegerHyperparameter("b_int", 0, 3, q=1.0)
#         expected = ("b_int", 'param_0 = pyll.scope.int(hp.quniform('
#                     '"b_int", -0.49999, 3.5, 1.0))')
#         value = self.pyll_writer.write_hyperparameter(b_int_1, None)
#         self.assertEqual(expected, value)
#
#         # TODO: trying to add the same parameter name a second time, maybe an
#         #  error should be raised!
#         b_int_2 = UniformIntegerHyperparameter("b_int", 0, 3, q=2.0)
#         expected = ("b_int", 'param_1 = pyll.scope.int(hp.quniform('
#                     '"b_int", -0.49999, 3.5, 2.0))')
#         value = self.pyll_writer.write_hyperparameter(b_int_2, None)
#         self.assertEqual(expected, value)
#
#         b_int_3 = UniformIntegerHyperparameter( "b_int", 1, 100, base=10)
#         b_int_3.name = self.pyll_writer.convert_name(b_int_3)
#         # TODO: this is an example of non-uniform integer sampling!
#         expected = ('LOG10_Q1_b_int', 'param_1 = hp.uniform('
#                     '"LOG10_Q1_b_int", -0.301021309861, 2.00216606176)')
#         value = self.pyll_writer.write_hyperparameter(b_int_3, None)
#         self.assertEqual(expected, value)
#
#     def test_write_loguniform(self):
#         c = UniformFloatHyperparameter("c", 0.001, 1, base=np.e)
#         expected = ("c", 'param_0 = hp.loguniform("c", -6.90775527898, 0.0)')
#         value = self.pyll_writer.write_hyperparameter(c, None)
#         self.assertEqual(expected, value)
#
#     def test_write_loguniform_int(self):
#         c_int = UniformIntegerHyperparameter("c_int", 1, 10, base=np.e)
#         expected = ("c_int", 'param_0 = pyll.scope.int(hp.qloguniform('
#                     '"c_int", -0.69312718076, 2.35137525716, 1.0))')
#         value = self.pyll_writer.write_hyperparameter(c_int, None)
#         self.assertEqual(expected, value)
#
#     def test_write_qloguniform(self):
#         d = UniformFloatHyperparameter("d", 0.1, 3, q=0.1, base=np.e)
#         expected = ("d", 'param_0 = hp.qloguniform("d", -2.99373427089, '
#                          '1.11514159062, 0.1)')
#         value = self.pyll_writer.write_hyperparameter(d, None)
#         self.assertEqual(expected, value)
#
#     def test_write_qloguniform_int(self):
#         d_int_1 = UniformIntegerHyperparameter("d_int", 1, 3, q=1.0, base=np.e)
#         expected = ("d_int", 'param_0 = pyll.scope.int(hp.qloguniform('
#                     '"d_int", -0.69312718076, 1.2527629685, 1.0))')
#         value = self.pyll_writer.write_hyperparameter(d_int_1, None)
#         self.assertEqual(expected, value)
#
#         d_int_2 = UniformIntegerHyperparameter("d_int", 1, 3, q=2.0, base=np.e)
#         expected = ("d_int", 'param_1 = pyll.scope.int(hp.qloguniform('
#                     '"d_int", -0.69312718076, 1.2527629685, 2.0))')
#         value = self.pyll_writer.write_hyperparameter(d_int_2, None)
#         self.assertEqual(expected, value)
#
#     def test_write_normal(self):
#         parameter = NormalFloatHyperparameter("e", 0, 1)
#         expected = ('e', 'param_0 = hp.normal("e", 0.0, 1.0)')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#         parameter = NormalFloatHyperparameter("e", 0, 1, base=10)
#         parameter.name = self.pyll_writer.convert_name(parameter)
#         expected = ('LOG10_e', 'param_1 = hp.normal("LOG10_e", 0.0, 1.0)')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#     def test_write_normal_int(self):
#         parameter = NormalIntegerHyperparameter("e", 0, 1)
#         expected = ('e',
#                     'param_0 = pyll.scope.int(hp.qnormal("e", 0.0, 1.0, 1.0))')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#         parameter = NormalIntegerHyperparameter("e", 0, 1, base=10)
#         parameter.name = self.pyll_writer.convert_name(parameter)
#         # TODO: this is an example of non-uniform sampling
#         expected = ('LOG10_Q1_e', 'param_1 = hp.normal("LOG10_Q1_e", 0.0, 1.0)')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#     def test_write_qnormal(self):
#         parameter = NormalFloatHyperparameter("f", 0, 1, q=0.1)
#         expected = ('f', 'param_0 = hp.qnormal("f", 0.0, 1.0, 0.1)')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#         parameter = NormalFloatHyperparameter("f", 0, 1, q=0.1, base=10)
#         parameter.name = self.pyll_writer.convert_name(parameter)
#         expected = ('LOG10_Q0.100000_f',
#                     'param_1 = hp.normal("LOG10_Q0.100000_f", 0.0, 1.0)')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#     def test_write_qnormal_int(self):
#         parameter = NormalIntegerHyperparameter("f_int", 0, 1, q=1.0)
#         expected = ('f_int',
#                     'param_0 = pyll.scope.int(hp.qnormal("f_int", 0.0, 1.0, 1.0))')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#         parameter = NormalIntegerHyperparameter("f_int", 0, 1, q=1.0, base=10)
#         parameter.name = self.pyll_writer.convert_name(parameter)
#         expected = ('LOG10_Q1.000000_f_int',
#                     'param_1 = hp.normal("LOG10_Q1.000000_f_int", 0.0, 1.0)')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#     def test_write_lognormal(self):
#         parameter = NormalFloatHyperparameter("g", 0, 1, base=np.e)
#         expected = ('g', 'param_0 = hp.lognormal("g", 0.0, 1.0)')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#     def test_write_lognormal_int(self):
#         parameter = NormalIntegerHyperparameter("g", 0, 1, base=np.e)
#         expected = ('g',
#                     'param_0 = pyll.scope.int(hp.qlognormal("g", 0.0, 1.0, 1.0))')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#     def test_write_qlognormal(self):
#         parameter = NormalFloatHyperparameter("g", 0, 1, q=0.1, base=np.e)
#         expected = ('g', 'param_0 = hp.qlognormal("g", 0.0, 1.0, 0.1)')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#     def test_write_qlognormal_int(self):
#         parameter = NormalIntegerHyperparameter("g_int", 0, 10, q=2.0, base=np.e)
#         expected = ('g_int',
#                     'param_0 = pyll.scope.int(hp.qlognormal("g_int", 0.0, 10.0, 2.0))')
#         value = self.pyll_writer.write_hyperparameter(parameter, None)
#         self.assertEqual(expected, value)
#
#     def test_get_bounds_as_exponent(self):
#         parameter = UniformFloatHyperparameter('a', 1, 1000, base=10)
#         lower, upper = self.pyll_writer.get_bounds_as_exponent(parameter)
#         name = self.pyll_writer.convert_name(parameter)
#         self.assertEqual(name, 'LOG10_a')
#         self.assertEqual(lower, 0)
#         self.assertEqual(upper, 3)
#
#         parameter = UniformFloatHyperparameter('a', 2, 128, base=2)
#         name = self.pyll_writer.convert_name(parameter)
#         lower, upper = self.pyll_writer.get_bounds_as_exponent(parameter)
#         self.assertEqual(name, 'LOG2_a')
#         self.assertEqual(lower, 1)
#         self.assertEqual(upper, 7)
#
#         parameter = UniformFloatHyperparameter('a', 128, 256, base=np.e)
#         name = self.pyll_writer.convert_name(parameter)
#         lower, upper = self.pyll_writer.get_bounds_as_exponent(parameter)
#         self.assertEqual(name, 'LOG_a')
#         self.assertAlmostEqual(lower, 4.852030264)
#         self.assertAlmostEqual(upper, 5.545177444)
#
#         parameter = UniformFloatHyperparameter('a', 10, 1000, base=5)
#         lower, upper = self.pyll_writer.get_bounds_as_exponent(parameter)
#         name = self.pyll_writer.convert_name(parameter)
#         self.assertEqual(name, 'LOG5_a')
#         self.assertAlmostEqual(lower, 1.430676558)
#         self.assertAlmostEqual(upper, 4.292029674)
#
#         parameter = UniformFloatHyperparameter('illegal',
#             0, 1000, base=10)
#         self.assertRaises(ValueError, self.pyll_writer.get_bounds_as_exponent,
#                           parameter)


# Copyright (c) 2014-2016, ConfigSpace developers
# Matthias Feurer
# Katharina Eggensperger
# and others (see commit history).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from collections import defaultdict, deque, OrderedDict
import copy
from itertools import chain

import numpy as np
import io

import ConfigSpace.nx
from ConfigSpace.hyperparameters import Hyperparameter, Constant, FloatHyperparameter
from ConfigSpace.conditions import ConditionComponent, \
    AbstractCondition, AbstractConjunction, EqualsCondition
# from ConfigSpace.forbidden import AbstractForbiddenComponent
from ConfigSpace.forbidden import AbstractForbiddenComponent
from typing import Union, List, Any, Dict, Iterable, Set, Tuple, Optional
from ConfigSpace.exceptions import ForbiddenValueError
import ConfigSpace.c_util


class ConfigurationSpace(object):
    # TODO add comments to both the configuration space and single
    # hyperparameters!

    # TODO add a method to add whole configuration spaces as a child "tree"

    """Represent a configuration space.
    """

    def __init__(
            self,
            name: Union[str, None] = None,
            seed: Union[int, None] = None,
            meta: Optional[Dict] = None,
    ) -> None:
        self.name = name
        self.meta = meta

        self._hyperparameters = OrderedDict()  # type: OrderedDict[str, Hyperparameter]
        self._hyperparameter_idx = dict()  # type: Dict[str, int]
        self._idx_to_hyperparameter = dict()  # type: Dict[int, str]

        # Use dictionaries to make sure that we don't accidently add
        # additional keys to these mappings (which happened with defaultdict()).
        # This once broke auto-sklearn's equal comparison of configuration
        # spaces when _children of one instance contained  all possible
        # hyperparameters as keys and empty dictionaries as values while the
        # other instance not containing these.
        self._children = OrderedDict()   # type: OrderedDict[str, OrderedDict[str, Union[None, AbstractCondition]]]
        self._parents = OrderedDict()   # type: OrderedDict[str, OrderedDict[str, Union[None, AbstractCondition]]]

        # changing this to a normal dict will break sampling because there is
        #  no guarantee that the parent of a condition was evaluated before
        self._conditionals = set()   # type: Set[str]
        self.forbidden_clauses = []  # type: List['AbstractForbiddenComponent']
        self.random = np.random.RandomState(seed)

        self._children['__HPOlib_configuration_space_root__'] = OrderedDict()

        # caching
        self._parent_conditions_of = dict()
        self._child_conditions_of = dict()
        self._parents_of = dict()
        self._children_of = dict()

    def generate_all_continuous_from_bounds(self, bounds: List[List[Any]]) -> None:
        for i, (l, u) in enumerate(bounds):
            hp = ConfigSpace.UniformFloatHyperparameter('x%d' % i, l, u)
            self.add_hyperparameter(hp)

    def add_hyperparameters(self, hyperparameters: List[Hyperparameter]) -> List[Hyperparameter]:
        """Add hyperparameters to the configuration space.

        Parameters
        ----------
        hyperparameters : list
            List of hyperparameters to add.

        Returns
        -------
        hyperparameters : list
            List of added hyperparameters (same as input)
        """
        for hyperparameter in hyperparameters:
            if not isinstance(hyperparameter, Hyperparameter):
                raise TypeError("Hyperparameter '%s' is not an instance of "
                                "ConfigSpace.hyperparameters.Hyperparameter." %
                                str(hyperparameter))

        for hyperparameter in hyperparameters:
            self._add_hyperparameter(hyperparameter)

        self._update_cache()
        self._check_default_configuration()
        self._sort_hyperparameters()
        return hyperparameters


    def add_hyperparameter(self, hyperparameter: Hyperparameter) -> Hyperparameter:
        """Add a hyperparameter to the configuration space.

        Parameters
        ----------
        hyperparameter : :class:`HPOlibConfigSpace.hyperparameters.
                Hyperparameter`
            The hyperparameter to add.
        """
        if not isinstance(hyperparameter, Hyperparameter):
            raise TypeError("The method add_hyperparameter must be called "
                            "with an instance of "
                            "ConfigSpace.hyperparameters.Hyperparameter.")

        self._add_hyperparameter(hyperparameter)
        self._update_cache()
        self._check_default_configuration()
        self._sort_hyperparameters()

        return hyperparameter

    def _add_hyperparameter(self, hyperparameter: Hyperparameter) -> None:
        # Check if adding the hyperparameter is legal:
        # * Its name must not already exist
        if hyperparameter.name in self._hyperparameters:
            raise ValueError("Hyperparameter '%s' is already in the "
                             "configuration space." % hyperparameter.name)
        self._hyperparameters[hyperparameter.name] = hyperparameter
        self._children[hyperparameter.name] = OrderedDict()

        # TODO remove __HPOlib_configuration_space_root__, it is only used in
        # to check for cyclic configuration spaces. If it is only added when
        # cycles are checked, the code can become much easier (e.g. the parent
        # caching can be more or less removed).
        self._children['__HPOlib_configuration_space_root__'][
            hyperparameter.name] = None
        self._parents[hyperparameter.name] = OrderedDict()
        self._parents[hyperparameter.name][
            '__HPOlib_configuration_space_root__'] = None
        # Save the index of each hyperparameter name to later on access a
        # vector of hyperparameter values by indices, must be done twice
        # because check_default_configuration depends on it
        for i, hp in enumerate(self._hyperparameters):
            self._hyperparameter_idx[hp] = i

    def add_condition(self, condition: ConditionComponent) -> ConditionComponent:
        # Check if adding the condition is legal:
        # * The parent in a condition statement must exist
        # * The condition must add no cycles
        # The following array keeps track of all edges which must be
        # added to the DiGraph; if the checks don't raise any Exception,
        # these edges are finally added at the end of the function
        if not isinstance(condition, ConditionComponent):
            raise TypeError("The method add_condition must be called "
                            "with an instance of "
                            "ConfigSpace.condition.ConditionComponent.")

        if isinstance(condition, AbstractCondition):
            self._check_edges(
                [(condition.parent, condition.child)],
                [condition.value],
            )
            self._check_condition(condition.child, condition)
            self._add_edge(
                condition.parent,
                condition.child,
                condition,
            )

        # Loop over the Conjunctions to find out the conditions we must add!
        elif isinstance(condition, AbstractConjunction):
            dlcs = condition.get_descendant_literal_conditions()
            edges = [(dlc.parent, dlc.child) for dlc in dlcs]
            values = [dlc.value for dlc in dlcs]
            self._check_edges(edges, values)

            for dlc in dlcs:
                self._check_condition(dlc.child, condition)
                self._add_edge(
                    dlc.parent,
                    dlc.child,
                    condition=condition,
                )

        else:
            raise Exception("This should never happen!")

        self._sort_hyperparameters()
        self._update_cache()
        return condition

    def add_conditions(self, conditions: List[ConditionComponent]) -> List[ConditionComponent]:
        for condition in conditions:
            if not isinstance(condition, ConditionComponent):
                raise TypeError("Condition '%s' is not an instance of "
                                "ConfigSpace.condition.ConditionComponent." %
                                str(condition))

        edges = []
        values = []
        conditions_to_add = []
        for condition in conditions:
            if isinstance(condition, AbstractCondition):
                edges.append((condition.parent, condition.child))
                values.append(condition.value)
                conditions_to_add.append(condition)
            elif isinstance(condition, AbstractConjunction):
                dlcs = condition.get_descendant_literal_conditions()
                edges.extend(
                    [(dlc.parent, dlc.child) for dlc in dlcs])
                values.extend([dlc.value for dlc in dlcs])
                conditions_to_add.extend([condition] * len(dlcs))

        for edge, condition in zip(edges, conditions_to_add):
            self._check_condition(edge[1], condition)
        self._check_edges(edges, values)
        for edge, condition in zip(edges, conditions_to_add):
            self._add_edge(edge[0], edge[1], condition)

        self._sort_hyperparameters()
        self._update_cache()
        return conditions

    def _add_edge(
            self,
            parent_node: Hyperparameter,
            child_node: Hyperparameter,
            condition: ConditionComponent,
    ) -> None:
        try:
            # TODO maybe this has to be done more carefully
            del self._children['__HPOlib_configuration_space_root__'][child_node.name]
        except Exception:
            pass

        try:
            del self._parents[child_node.name]['__HPOlib_configuration_space_root__']
        except Exception:
            pass

        self._children[parent_node.name][child_node.name] = condition
        self._parents[child_node.name][parent_node.name] = condition
        self._conditionals.add(child_node.name)

    def _check_condition(self, child_node: Hyperparameter, condition: ConditionComponent) \
            -> None:
        for other_condition in self._get_parent_conditions_of(child_node.name):
            if other_condition != condition:
                raise ValueError("Adding a second condition (different) for a "
                                 "hyperparameter is ambigouos and "
                                 "therefore forbidden. Add a conjunction "
                                 "instead!\nAlready inserted: %s\nNew one: "
                                 "%s" % (str(other_condition), str(condition)))

    def _check_edges(
            self,
            edges: List[Tuple[Hyperparameter, Hyperparameter]],
            values: List[Union[float, str, int]]
    ) -> None:
        for (parent_node, child_node), value in zip(edges, values):
            # check if both nodes are already inserted into the graph
            if child_node.name not in self._hyperparameters:
                raise ValueError(
                    "Child hyperparameter '%s' not in configuration "
                    "space." % child_node.name)
            if child_node != self._hyperparameters[child_node.name]:
                # TODO test this
                raise ValueError(
                    "Child hyperparameter '%s' different to hyperparameter "
                    "with the same name in configuration space: '%s'." %
                    (child_node, self._hyperparameters[child_node.name])
                )
            if parent_node.name not in self._hyperparameters:
                raise ValueError(
                    "Parent hyperparameter '%s' not in configuration "
                    "space." % parent_node.name)
            if parent_node != self._hyperparameters[parent_node.name]:
                # TODO test this
                raise ValueError(
                    "Parent hyperparameter '%s' different to hyperparameter "
                    "with the same name in configuration space: '%s'." %
                    (parent_node, self._hyperparameters[parent_node.name])
                )
            if isinstance(value, (tuple, list)):
                # TODO test this
                for v in value:
                    if not self._hyperparameters[parent_node.name].is_legal(v):
                        raise ValueError(
                            "Value '%s' is not legal for hyperparameter %s." %
                            (v, self._hyperparameters[parent_node.name])
                        )
            else:
                if not self._hyperparameters[parent_node.name].is_legal(value):
                    raise ValueError(
                        "Value '%s' is not legal for hyperparameter %s." %
                        (value, self._hyperparameters[parent_node.name])
                    )

        # TODO: recursively check everything which is inside the conditions,
        # this means we have to recursively traverse the condition

        tmp_dag = self._create_tmp_dag()
        for parent_node, child_node in edges:
            tmp_dag.add_edge(parent_node.name, child_node.name)

        if not ConfigSpace.nx.is_directed_acyclic_graph(tmp_dag):
            cycles = list(
                ConfigSpace.nx.simple_cycles(tmp_dag)
            )  # type: List[List[str]]
            for cycle in cycles:
                cycle.sort()
            cycles.sort()
            raise ValueError("Hyperparameter configuration contains a "
                             "cycle %s" % str(cycles))

    def _sort_hyperparameters(self) -> None:
        levels = OrderedDict()  # type: OrderedDict[str, int]
        to_visit = deque() # type: ignore
        for hp_name in self._hyperparameters:
            to_visit.appendleft(hp_name)

        while len(to_visit) > 0:
            current = to_visit.pop()
            if '__HPOlib_configuration_space_root__' in self._parents[current]:
                assert len(self._parents[current]) == 1
                levels[current] = 1

            else:
                all_parents_visited = True
                depth = -1
                for parent in self._parents[current]:
                    if parent not in levels:
                        all_parents_visited = False
                        break
                    else:
                        depth = max(depth, levels[parent] + 1)

                if all_parents_visited:
                    levels[current] = depth
                else:
                    to_visit.appendleft(current)

        by_level = defaultdict(list)  # type: defaultdict[int, List[str]]
        for hp in levels:
            level = levels[hp]
            by_level[level].append(hp)

        nodes = []
        # Sort and add to list
        for level in sorted(by_level):
            sorted_by_level = by_level[level]
            sorted_by_level.sort()
            nodes.extend(sorted_by_level)

        # Resort the OrderedDict
        new_order = OrderedDict()
        for node in nodes:
            new_order[node] = self._hyperparameters[node]
        self._hyperparameters = new_order

        # Update to reflect sorting
        for i, hp in enumerate(self._hyperparameters):
            self._hyperparameter_idx[hp] = i
            self._idx_to_hyperparameter[i] = hp

        # Update order of _children
        new_order = OrderedDict()
        new_order['__HPOlib_configuration_space_root__'] = self._children['__HPOlib_configuration_space_root__']
        for hp in chain(['__HPOlib_configuration_space_root__'], self._hyperparameters):
            # Also resort the children dict
            children_sorting = [(self._hyperparameter_idx[child_name], child_name)
                                for child_name in self._children[hp]]
            children_sorting.sort()
            children_order = OrderedDict()
            for _, child_name in children_sorting:
                children_order[child_name] = self._children[hp][child_name]
            new_order[hp] = children_order
        self._children = new_order

        # Update order of _parents
        new_order = OrderedDict()
        for hp in self._hyperparameters:
            # Also resort the parent's dict
            if '__HPOlib_configuration_space_root__' in self._parents[hp]:
                parent_sorting = [(-1, '__HPOlib_configuration_space_root__')]
            else:
                parent_sorting = [(self._hyperparameter_idx[parent_name], parent_name)
                                  for parent_name in self._parents[hp]]
            parent_sorting.sort()
            parent_order = OrderedDict()
            for _, parent_name in parent_sorting:
                parent_order[parent_name] = self._parents[hp][parent_name]
            new_order[hp] = parent_order
        self._parents = new_order

        # update conditions
        for condition in self.get_conditions():
            condition.set_vector_idx(self._hyperparameter_idx)

        # forbidden clauses
        for clause in self.get_forbiddens():
            clause.set_vector_idx(self._hyperparameter_idx)

    def _update_cache(self):
        self._parent_conditions_of = dict()
        self._child_conditions_of = dict()
        self._parents_of = dict()
        self._children_of = dict()

        for hp_name in self._hyperparameters:
            self._parent_conditions_of[hp_name] = self._get_parent_conditions_of(hp_name)
            self._child_conditions_of[hp_name] = self._get_child_conditions_of(hp_name)
            self._parents_of[hp_name] = self.get_parents_of(hp_name)
            self._children_of[hp_name] = self.get_children_of(hp_name)

    def _create_tmp_dag(self) -> ConfigSpace.nx.DiGraph:
        tmp_dag = ConfigSpace.nx.DiGraph()
        for hp_name in self._hyperparameters:
            tmp_dag.add_node(hp_name)
            tmp_dag.add_edge('__HPOlib_configuration_space_root__', hp_name)

        for parent_node_ in self._children:
            if parent_node_ == '__HPOlib_configuration_space_root__':
                continue
            for child_node_ in self._children[parent_node_]:
                try:
                    tmp_dag.remove_edge('__HPOlib_configuration_space_root__',
                                        child_node_)
                except Exception:
                    pass
                condition = self._children[parent_node_][child_node_]
                tmp_dag.add_edge(parent_node_, child_node_, condition=condition)

        return tmp_dag

    def add_forbidden_clause(self, clause: AbstractForbiddenComponent) -> AbstractForbiddenComponent:
        if not isinstance(clause, AbstractForbiddenComponent):
            raise TypeError("The method add_forbidden_clause must be called "
                            "with an instance of "
                            "ConfigSpace.forbidden.AbstractForbiddenComponent.")
        clause.set_vector_idx(self._hyperparameter_idx)
        self.forbidden_clauses.append(clause)
        self._check_default_configuration()
        return clause

    def add_forbidden_clauses(self, clauses: List[AbstractForbiddenComponent]) -> List[AbstractForbiddenComponent]:
        for clause in clauses:
            clause.set_vector_idx(self._hyperparameter_idx)
            if not isinstance(clause, AbstractForbiddenComponent):
                raise TypeError("Forbidden '%s' is not an instance of "
                                "ConfigSpace.forbidden.AbstractForbiddenComponent." %
                                str(clause))
            self.forbidden_clauses.append(clause)
        self._check_default_configuration()
        return clauses

    def add_configuration_space(self, prefix: str, configuration_space: 'ConfigurationSpace',
                                delimiter: str=":", parent_hyperparameter: Hyperparameter=None) -> 'ConfigurationSpace':
        if not isinstance(configuration_space, ConfigurationSpace):
            raise TypeError("The method add_configuration_space must be "
                            "called with an instance of "
                            "HPOlibConfigSpace.configuration_space."
                            "ConfigurationSpace.")

        new_parameters = []
        for hp in configuration_space.get_hyperparameters():
            new_parameter = copy.copy(hp)
            # Allow for an empty top-level parameter
            if new_parameter.name == '':
                new_parameter.name = prefix
            else:
                new_parameter.name = "%s%s%s" % (prefix, delimiter,
                                                 new_parameter.name)
            new_parameters.append(new_parameter)
        self.add_hyperparameters(new_parameters)

        conditions_to_add = []
        for condition in configuration_space.get_conditions():
            new_condition = copy.copy(condition)
            dlcs = new_condition.get_descendant_literal_conditions()
            for dlc in dlcs:
                if dlc.child.name == prefix or dlc.child.name == '':
                    dlc.child.name = prefix
                elif not dlc.child.name.startswith(
                                "%s%s" % (prefix, delimiter)):
                    dlc.child.name = "%s%s%s" % (
                        prefix, delimiter, dlc.child.name)
                if dlc.parent.name == prefix or dlc.parent.name == '':
                    dlc.parent.name = prefix
                elif not dlc.parent.name.startswith(
                                "%s%s" % (prefix, delimiter)):
                    dlc.parent.name = "%s%s%s" % (
                        prefix, delimiter, dlc.parent.name)
            conditions_to_add.append(new_condition)
        self.add_conditions(conditions_to_add)

        forbiddens_to_add = []
        for forbidden_clause in configuration_space.forbidden_clauses:
            # new_forbidden = copy.deepcopy(forbidden_clause)
            new_forbidden = forbidden_clause
            dlcs = new_forbidden.get_descendant_literal_clauses()
            for dlc in dlcs:
                if dlc.hyperparameter.name == prefix or \
                                dlc.hyperparameter.name == '':
                    dlc.hyperparameter.name = prefix
                elif not dlc.hyperparameter.name.startswith(
                                "%s%s" % (prefix, delimiter)):
                    dlc.hyperparameter.name = "%s%s%s" % \
                                              (prefix, delimiter,
                                               dlc.hyperparameter.name)
            forbiddens_to_add.append(new_forbidden)
        self.add_forbidden_clauses(forbiddens_to_add)

        conditions_to_add = []
        if parent_hyperparameter is not None:
            for new_parameter in new_parameters:
                # Only add a condition if the parameter is a top-level
                # parameter of the new configuration space (this will be some
                #  kind of tree structure).
                if self.get_parents_of(new_parameter):
                    continue
                condition = EqualsCondition(new_parameter,
                                            parent_hyperparameter['parent'],
                                            parent_hyperparameter['value'])
                conditions_to_add.append(condition)
        self.add_conditions(conditions_to_add)

        return configuration_space

    def get_hyperparameters(self) -> List[Hyperparameter]:
        return list(self._hyperparameters.values())

    def get_hyperparameter_names(self) -> List[str]:
        return list(self._hyperparameters.keys())

    def get_hyperparameter(self, name: str) -> Hyperparameter:
        hp = self._hyperparameters.get(name)

        if hp is None:
            if self.name is None:
                raise KeyError("Hyperparameter '%s' does not exist in this "
                               "configuration space." % name)
            else:
                raise KeyError("Hyperparameter '%s' does not exist in  "
                               "configuration space %s." % (name, self.name))
        else:
            return hp

    def get_hyperparameter_by_idx(self, idx: int) -> str:
        hp = self._idx_to_hyperparameter.get(idx)

        if hp is None:
            if self.name is None:
                raise KeyError("Hyperparameter #'%d' does not exist in this "
                               "configuration space." % idx)
            else:
                raise KeyError("Hyperparameter #'%d' does not exist in  "
                               "configuration space %s." % (idx, self.name))
        else:
            return hp

    def get_idx_by_hyperparameter_name(self, name: str) -> int:
        idx = self._hyperparameter_idx.get(name)

        if idx is None:
            if self.name is None:
                raise KeyError("Hyperparameter '%s' does not exist in this "
                               "configuration space." % name)
            else:
                raise KeyError("Hyperparameter '%s' does not exist in  "
                               "configuration space %s." % (name, self.name))
        else:
            return idx

    def get_conditions(self) -> List[AbstractCondition]:
        conditions = []
        added_conditions = set()  # type: Set[str]

        # Nodes is a list of nodes
        for source_node in self.get_hyperparameters():
            # This is a list of keys in a dictionary
            # TODO sort the edges by the order of their source_node in the
            # hyperparameter list!
            for target_node in self._children[source_node.name]:
                if target_node not in added_conditions:
                    condition = self._children[source_node.name][target_node]
                    conditions.append(condition)
                    added_conditions.add(target_node)

        return conditions

    def get_forbiddens(self) -> List[AbstractForbiddenComponent]:
        return self.forbidden_clauses

    def get_children_of(self, name: Union[str, Hyperparameter]) -> List[Hyperparameter]:
        conditions = self.get_child_conditions_of(name)
        parents = [] # type: List[Hyperparameter]
        for condition in conditions:
            parents.extend(condition.get_children())
        return parents

    def _get_children_of(self, name: str) -> List[Hyperparameter]:
        conditions = self._get_child_conditions_of(name)
        parents = []  # type: List[Hyperparameter]
        for condition in conditions:
            parents.extend(condition.get_children())
        return parents

    def get_child_conditions_of(self, name: Union[str, Hyperparameter]) -> List[AbstractCondition]:
        if isinstance(name, Hyperparameter):
            name = name.name  # type: ignore

        # This raises an exception if the hyperparameter does not exist
        self.get_hyperparameter(name)
        return self._get_child_conditions_of(name)

    def _get_child_conditions_of(self, name: str) -> List[AbstractCondition]:
        children = self._children[name]
        conditions = [children[child_name] for child_name in children
                      if child_name != "__HPOlib_configuration_space_root__"]
        return conditions

    def get_parents_of(self, name: Union[str, Hyperparameter]) -> List[Hyperparameter]:
        """Return the parent hyperparameters of a given hyperparameter.

        Parameters
        ----------
        name : str or Hyperparameter
            Can either be the name of a hyperparameter or the hyperparameter
            object.

        Returns
        -------
        list
            List with all parent hyperparameters.
        """
        conditions = self.get_parent_conditions_of(name)
        parents = [] # type: List[Hyperparameter]
        for condition in conditions:
            parents.extend(condition.get_parents())
        return parents

    def _get_parents_of(self, name: str) -> List[Hyperparameter]:
        """Return the parent hyperparameters of a given hyperparameter.

        Parameters
        ----------
        name : str

        Returns
        -------
        list
            List with all parent hyperparameters.
        """
        conditions = self._get_parent_conditions_of(name)
        parents = []  # type: List[Hyperparameter]
        for condition in conditions:
            parents.extend(condition.get_parents())
        return parents

    def get_parent_conditions_of(self, name: Union[str, Hyperparameter]) -> List[AbstractCondition]:
        if isinstance(name, Hyperparameter):
            name = name.name  # type: ignore

        # This raises an exception if the hyperparameter does not exist
        self.get_hyperparameter(name)
        return self._get_parent_conditions_of(name)

    def _get_parent_conditions_of(self, name: str) -> List[AbstractCondition]:
        parents = self._parents[name]
        conditions = [parents[parent_name] for parent_name in parents
                      if parent_name != "__HPOlib_configuration_space_root__"]
        return conditions

    def get_all_unconditional_hyperparameters(self) -> List[str]:
        hyperparameters = [hp_name for hp_name in
                           self._children[
                               '__HPOlib_configuration_space_root__']]
        return hyperparameters

    def get_all_conditional_hyperparameters(self) -> List[str]:
        return self._conditionals

    def get_default_configuration(self) -> 'Configuration':
        return self._check_default_configuration()

    def _check_default_configuration(self) -> 'Configuration':
        # Check if adding that hyperparameter leads to an illegal default configuration
        instantiated_hyperparameters = {} # type: Dict[str, Optional[Union[int, float, str]]]
        for hp in self.get_hyperparameters():
            conditions = self._get_parent_conditions_of(hp.name)
            active = True
            for condition in conditions:
                parent_names = [c.parent.name for c in
                                condition.get_descendant_literal_conditions()]

                parents = {
                    parent_name: instantiated_hyperparameters[parent_name]
                    for parent_name in parent_names
                }

                if not condition.evaluate(parents):
                    # TODO find out why a configuration is illegal!
                    active = False

            if active == False:
                instantiated_hyperparameters[hp.name] = None
            elif isinstance(hp, Constant):
                instantiated_hyperparameters[hp.name] = hp.value
            else:
                instantiated_hyperparameters[hp.name] = hp.default_value

                # TODO copy paste from check configuration

        # TODO add an extra Exception type for the case that the default
        # configuration is forbidden!
        return Configuration(self, instantiated_hyperparameters)

    # For backward compatibility
    def check_configuration(self, configuration: 'Configuration') -> None:
        if not isinstance(configuration, Configuration):
            raise TypeError("The method check_configuration must be called "
                            "with an instance of %s. "
                            "Your input was of type %s"% (Configuration, type(configuration)))
        ConfigSpace.c_util.check_configuration(
            self, configuration.get_array(), False
        )

    def check_configuration_vector_representation(self, vector: np.ndarray) -> None:
        if not isinstance(vector, np.ndarray):
            raise TypeError("The method check_configuration must be called "
                            "with an instance of np.ndarray "
                            "Your input was of type %s" % (type(vector)))
        ConfigSpace.c_util.check_configuration(self, vector, False)

    def get_active_hyperparameters(self, configuration: 'Configuration') -> Set:
        vector = configuration.get_array()
        active_hyperparameters = set()
        for hp_name, hyperparameter in self._hyperparameters.items():
            conditions = self._parent_conditions_of[hyperparameter.name]

            active = True
            for condition in conditions:

                parent_vector_idx = condition.get_parents_vector()

                # if one of the parents is None, the hyperparameter cannot be
                # active! Else we have to check this
                # Note from trying to optimize this - this is faster than using
                # dedicated numpy functions and indexing
                if any([vector[i] != vector[i] for i in parent_vector_idx]):
                    active = False
                    break

                else:
                    if not condition.evaluate_vector(vector):
                        active = False
                        break

            if active:
                active_hyperparameters.add(hp_name)
        return active_hyperparameters

    def _check_configuration_rigorous(self, configuration: 'Configuration',
                                      allow_inactive_with_values: bool = False) -> None:
        vector = configuration.get_array()
        active_hyperparameters = self.get_active_hyperparameters(configuration)

        for hp_name, hyperparameter in self._hyperparameters.items():
            hp_value = vector[self._hyperparameter_idx[hp_name]]
            active = hp_name in active_hyperparameters

            if not np.isnan(hp_value) and not hyperparameter.is_legal_vector(hp_value):
                raise ValueError("Hyperparameter instantiation '%s' "
                                 "(type: %s) is illegal for hyperparameter %s" %
                                 (hp_value, str(type(hp_value)),
                                  hyperparameter))

            if active and np.isnan(hp_value):
                raise ValueError("Active hyperparameter '%s' not specified!" %
                                 hyperparameter.name)

            if not allow_inactive_with_values and not active and \
                            not np.isnan(hp_value):
                raise ValueError("Inactive hyperparameter '%s' must not be "
                                 "specified, but has the vector value: '%s'." %
                                 (hp_name, hp_value))
        self._check_forbidden(vector)

    def _check_forbidden(self, vector: np.ndarray) -> None:
        ConfigSpace.c_util.check_forbidden(self.forbidden_clauses, vector)
        #for clause in self.forbidden_clauses:
        #    if clause.is_forbidden_vector(vector, strict=False):
        #        raise ForbiddenValueError("Given vector violates forbidden
    # clause %s" % (str(clause)))

    # http://stackoverflow.com/a/25176504/4636294
    def __eq__(self, other: Any) -> bool:
        """Override the default Equals behavior"""
        if isinstance(other, self.__class__):
            this_dict = self.__dict__.copy()
            del this_dict['random']
            other_dict = other.__dict__.copy()
            del other_dict['random']
            return this_dict == other_dict
        return NotImplemented

    def __ne__(self, other: Any) -> bool:
        """Define a non-equality test"""
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self) -> int:
        """Override the default hash behavior (that returns the id or the object)"""
        return hash(self.__repr__())

    def __repr__(self) -> str:
        retval = io.StringIO()
        retval.write("Configuration space object:\n  Hyperparameters:\n")

        if self.name is not None:
            retval.write(self.name)
            retval.write('\n')

        hyperparameters = sorted(self.get_hyperparameters(),
                                 key=lambda t: t.name)
        if hyperparameters:
            retval.write("    ")
            retval.write("\n    ".join(
                [str(hyperparameter) for hyperparameter in hyperparameters]))
            retval.write("\n")

        conditions = sorted(self.get_conditions(),
                            key=lambda t: str(t))
        if conditions:
            retval.write("  Conditions:\n")
            retval.write("    ")
            retval.write("\n    ".join(
                [str(condition) for condition in conditions]))
            retval.write("\n")

        if self.get_forbiddens():
            retval.write("  Forbidden Clauses:\n")
            retval.write("    ")
            retval.write("\n    ".join(
                [str(clause) for clause in self.get_forbiddens()]))
            retval.write("\n")

        retval.seek(0)
        return retval.getvalue()

    def __iter__(self) -> Iterable:
        """ Allows to iterate over the hyperparameter names in (hopefully?) the right order."""
        return iter(self._hyperparameters.keys())

    def sample_configuration(self, size: int = 1) -> Union['Configuration', List['Configuration']]:
        if not isinstance(size, int):
            raise TypeError('Argument size must be of type int, but is %s'
                            % type(size))
        elif size < 1:
            return []

        iteration = 0
        missing = size
        accepted_configurations = []  # type: List['Configuration']
        num_hyperparameters = len(self._hyperparameters)

        unconditional_hyperparameters = self.get_all_unconditional_hyperparameters()
        hyperparameters_with_children = list()

        _forbidden_clauses_unconditionals = []
        _forbidden_clauses_conditionals = []
        for clause in self.get_forbiddens():
            based_on_conditionals = False
            for subclause in clause.get_descendant_literal_clauses():
                if subclause.hyperparameter.name not in unconditional_hyperparameters:
                    based_on_conditionals = True
                    break
            if based_on_conditionals:
                _forbidden_clauses_conditionals.append(clause)
            else:
                _forbidden_clauses_unconditionals.append(clause)

        for uhp in unconditional_hyperparameters:
            children = self._children_of[uhp]
            if len(children) > 0:
                hyperparameters_with_children.append(uhp)

        while len(accepted_configurations) < size:
            if missing != size:
                missing = int(1.1 * missing)
            vector = np.ndarray((missing, num_hyperparameters),
                                dtype=np.float64)

            for i, hp_name in enumerate(self._hyperparameters):
                hyperparameter = self._hyperparameters[hp_name]
                vector[:, i] = hyperparameter._sample(self.random, missing)

            for i in range(missing):
                try:
                    configuration = Configuration(
                        self,
                        vector=ConfigSpace.c_util.correct_sampled_array(
                            vector[i].copy(),
                            _forbidden_clauses_unconditionals,
                            _forbidden_clauses_conditionals,
                            hyperparameters_with_children,
                            num_hyperparameters,
                            unconditional_hyperparameters,
                            self._hyperparameter_idx,
                            self._parent_conditions_of,
                            self._parents_of,
                            self._children_of,
                        ))
                    accepted_configurations.append(configuration)
                except ForbiddenValueError as e:
                    iteration += 1

                    if iteration == size * 100:
                        raise ForbiddenValueError(
                            "Cannot sample valid configuration for "
                            "%s" % self)

            missing = size - len(accepted_configurations)

        if size <= 1:
            return accepted_configurations[0]
        else:
            return accepted_configurations

    def seed(self, seed: int) -> None:
        self.random = np.random.RandomState(seed)


class Configuration(object):
    def __init__(self, configuration_space: ConfigurationSpace, values: Union[None,  Dict[str, Union[str, float, int]]] = None,
                 vector: Union[None, np.ndarray]=None, allow_inactive_with_values: bool=False, origin: Any=None)\
            -> None:
        """A single configuration.

        Parameters
        ----------
        configuration_space : ConfigurationSpace
            The configuration space for this configuration

        values : dict
            A dictionary with pairs (hyperparameter_name, value), where value is
            a legal value of the hyperparameter in the above
            configuration_space.

        vector : np.ndarray
            A numpy array for efficient representation. Either values or
            vector has to be given.

        allow_inactive_with_values : bool (default=False)
            Whether an Exception will be raised if a value for an inactive
            hyperparameter is given. Default is to raise an Exception.
        """
        if not isinstance(configuration_space, ConfigurationSpace):
            raise TypeError("Configuration expects an instance of %s, "
                            "you provided '%s'" %
                            (ConfigurationSpace, type(configuration_space)))

        self.configuration_space = configuration_space
        self.allow_inactive_with_values = allow_inactive_with_values
        self._query_values = False
        self._num_hyperparameters = len(self.configuration_space._hyperparameters)
        self.origin = origin
        self._keys = None  # type: Union[None, List[str]]

        if values is not None and vector is not None:
            raise ValueError('Configuration specified both as dictionary and '
                             'vector, can only do one.')
        if values is not None:
            # Using cs._hyperparameters to iterate makes sure that the
            # hyperparameters in the configuration are sorted in the same way as
            # they are sorted in the configuration space
            self._values = dict()  # type: Dict[str, Union[str, float, int]]
            for key in configuration_space._hyperparameters:
                value = values.get(key)
                if value is None:
                    continue
                hyperparameter = configuration_space.get_hyperparameter(key)
                if not hyperparameter.is_legal(value):
                    raise ValueError("Trying to set illegal value '%s' for"
                                     "hyperparameter %s." % (str(value), hyperparameter))
                # Truncate the representation of the float to be of constant
                # length for a python version
                if isinstance(hyperparameter, FloatHyperparameter):
                    value = float(repr(value))

                self._values[key] = value

            for key in values:
                if key not in configuration_space._hyperparameters:
                    raise ValueError('Tried to specify unknown hyperparameter '
                                     '%s' % key)

            self._query_values = True
            self._vector = np.ndarray((self._num_hyperparameters,),
                                      dtype=np.float)

            # Populate the vector
            # TODO very unintuitive calls...
            for key in configuration_space._hyperparameters:
                self._vector[self.configuration_space._hyperparameter_idx[
                    key]] = self.configuration_space.get_hyperparameter(key). \
                    _inverse_transform(self[key])
            self.is_valid_configuration()


        elif vector is not None:
            self._values = dict()
            if not isinstance(vector, np.ndarray):
                vector = np.array(vector, dtype=float)
            if len(vector.shape) > 1:
                if len(vector.shape) == 2 and vector.shape[1] == 1:
                    vector = vector.flatten()
                else:
                    raise ValueError(
                        'Only 1d arrays can be converted to a Configuration, '
                        'you passed an array of shape %s.' % str(vector.shape)
                    )
            if len(vector) != len(self.configuration_space.get_hyperparameters()):
                raise ValueError(
                    'Expected array of length %d, got %d' %
                    (len(self.configuration_space.get_hyperparameters()), len(vector))
                )
            self._vector = vector
        else:
            raise ValueError('Configuration neither specified as dictionary '
                             'or vector.')

    def is_valid_configuration(self) -> None:
        ConfigSpace.c_util.check_configuration(
            self.configuration_space,
            self._vector,
            allow_inactive_with_values=self.allow_inactive_with_values
        )

    def __getitem__(self, item: str) -> Any:
        if self._query_values or item in self._values:
            return self._values.get(item)

        hyperparameter = self.configuration_space._hyperparameters[item]
        item_idx = self.configuration_space._hyperparameter_idx[item]

        if not np.isfinite(self._vector[item_idx]):
            raise KeyError()

        value = hyperparameter._transform(self._vector[item_idx])
        # Truncate the representation of the float to be of constant
        # length for a python version
        if isinstance(hyperparameter, FloatHyperparameter):
            value = float(repr(value))
        # TODO make everything faster, then it'll be possible to init all values
        # at the same time and use an OrderedDict instead of only a dict here to
        # support iterating that dict in the same order as the actual order of
        # hyperparameters
        self._values[item] = value
        return self._values[item]

    def get(self, item: str, default: Union[None, Any]=None) -> Union[None, Any]:
        try:
            return self[item]
        except:
            return default
            
    def __setitem__(self, key, value):
        param = self.configuration_space.get_hyperparameter(key)
        if not param.is_legal(value):
            raise ValueError(
                "Illegal value '%s' for hyperparameter %s" % (str(value), key))
        idx = self.configuration_space.get_idx_by_hyperparameter_name(key)
        vector_value = param._inverse_transform(value)
        new_array = ConfigSpace.c_util.change_hp_value(
            self.configuration_space,
            self.get_array().copy(),
            param.name,
            vector_value,
            idx
        )
        ConfigSpace.c_util.check_configuration(
            self.configuration_space,
            new_array,
            False
        )
        self._vector = new_array
        self._values = dict()
        self._query_values = False
        
    def __contains__(self, item: str) -> bool:
        self._populate_values()
        return item in self._values

    # http://stackoverflow.com/a/25176504/4636294
    def __eq__(self, other: Any) -> bool:
        """Override the default Equals behavior"""
        if isinstance(other, self.__class__):
            self._populate_values()
            other._populate_values()
            return self._values == other._values and \
                self.configuration_space == other.configuration_space
        return NotImplemented

    def __ne__(self, other: Any) -> bool:
        """Define a non-equality test"""
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self) -> int:
        """Override the default hash behavior (that returns the id or the object)"""
        self._populate_values()
        return hash(self.__repr__())

    def _populate_values(self) -> None:
        if self._query_values is False:
            for hyperparameter in self.configuration_space.get_hyperparameters():
                self.get(hyperparameter.name)
            self._query_values = True

    def __repr__(self) -> str:
        self._populate_values()

        representation = io.StringIO()
        representation.write("Configuration:\n")

        hyperparameters = self.configuration_space.get_hyperparameters()
        hyperparameters.sort(key=lambda t: t.name)
        for hyperparameter in hyperparameters:
            hp_name = hyperparameter.name
            if hp_name in self._values and self._values[hp_name] is not None:
                representation.write("  ")

                value = repr(self._values[hp_name])
                if isinstance(hyperparameter, Constant):
                    representation.write("%s, Constant: %s" % (hp_name, value))
                else:
                    representation.write("%s, Value: %s" % (hp_name, value))
                representation.write("\n")

        return representation.getvalue()

    def __iter__(self) -> Iterable:
        return iter(self.keys())

    def keys(self) -> List[str]:
        # Cache the keys to speed up the process of retrieving the keys
        if self._keys is None:
            keys = list(self.configuration_space._hyperparameters.keys())
            keys = [
                key for i, key in enumerate(keys) if
                    np.isfinite(self._vector[i])
            ]
            self._keys = keys
        return self._keys

    def get_dictionary(self) -> Dict[str, Union[str, float, int]]:
        self._populate_values()
        return self._values

    def get_array(self) -> np.ndarray:
        """
        Returns
        -------
        numpy.ndarray
            internal vector representation of the configuration. All
            continuous values are scaled between zero and one.
        """
        return self._vector

